from abc import ABC, abstractmethod

class AbstractMultistart(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def to_dtype(self, dtype):
        pass

'''
To determine grid of initial conditions:
- Calculate linear resonance peak for each frequency and create a grid around that peak thus for every frequency
- Calulate linear resonance peak and and create grid with that peak (-peak, +peak, n_points)

To determine initial conditions of segment shooting method:
- Warm start with that initial condition and divide the time interval into m segments
'''


import jax
import jax.numpy as jnp

def select_branches(
    t_max_disp: jax.Array, # (N_COARSE_DRIVING_FREQUENCIES, N_COARSE_DRIVING_AMPLITUDES, N_COARSE_INITIAL_DISPLACEMENTS, N_COARSE_INITIAL_VELOCITIES, n_modes)
    y_max_disp: jax.Array, # (N_COARSE_DRIVING_FREQUENCIES, N_COARSE_DRIVING_AMPLITUDES, N_COARSE_INITIAL_DISPLACEMENTS, N_COARSE_INITIAL_VELOCITIES, 2 * n_modes)
    sweep_direction: int,
) -> tuple[jax.Array, jax.Array]:

    n_coarse_freq, n_coarse_amp, n_init_disp, n_init_vel, n_modes = t_max_disp.shape
    n_state      = 2 * n_modes
    n_branches   = n_init_disp * n_init_vel

    t_branches = t_max_disp.reshape(n_coarse_freq, n_coarse_amp, n_branches, n_modes) # (n_coarse_freq, n_coarse_amp, n_branches, n_modes)
    y_branches = y_max_disp.reshape(n_coarse_freq, n_coarse_amp, n_branches, n_state) # (n_coarse_freq, n_coarse_amp, n_branches, n_modes * 2)

    is_fwd = sweep_direction == 1
    idx_fwd = jnp.arange(n_coarse_freq, dtype=int) # (n_coarse_freq,)
    idx_bwd = jnp.arange(n_coarse_freq - 1, -1, -1, dtype=int) # (n_coarse_freq,)

    def _pick_for_amplitude(t_seq, y_seq): # (n_coarse_freq, n_branches, ..)
        sweep_order = jax.lax.cond(is_fwd, lambda _: idx_fwd, lambda _: idx_bwd, operand=None)
        t_ord, y_ord = t_seq[sweep_order], y_seq[sweep_order]

        # start branch = state with min ‖y‖² at first frequency step
        idx0   = jnp.argmin(jnp.sum(y_ord[0] ** 2, axis=-1))
        y0     = y_ord[0, idx0]
        t0     = t_ord[0, idx0]

        def _step(prev_y, xy):
            t_cur, y_cur = xy
            dist  = jnp.sum((y_cur - prev_y) ** 2, axis=-1)
            idx   = jnp.argmin(dist)
            y_sel = y_cur[idx]
            t_sel = t_cur[idx]
            return y_sel, (t_sel, y_sel)

        _, (t_rest, y_rest) = jax.lax.scan(_step, y0, (t_ord[1:], y_ord[1:]))
        t_sel_ord = jnp.concatenate((t0[None, :], t_rest), axis=0)
        y_sel_ord = jnp.concatenate((y0[None, :], y_rest), axis=0)

        t_sel = jax.lax.cond(is_fwd, lambda _: t_sel_ord, lambda _: t_sel_ord[::-1], operand=None)
        y_sel = jax.lax.cond(is_fwd, lambda _: y_sel_ord, lambda _: y_sel_ord[::-1], operand=None)
        return t_sel, y_sel

    t_max_disp, y_max_disp = jax.vmap(_pick_for_amplitude, in_axes=(1, 1), out_axes=(1, 1))(t_branches, y_branches)

    return t_max_disp, y_max_disp


import jax
import diffrax
from abc import ABC, abstractmethod

from ..oscillator.abstract_oscillator import AbstractOscillator
from ..excitation.abstract_excitation import AbstractExcitation
from .. import constants as const

class AbstractSolver(ABC):
    def __init__(self):
        pass
    
    @abstractmethod
    def time_response(self,
            excitation: AbstractExcitation, 
            init_disp: jax.Array,  
            init_vel: jax.Array
        ) -> tuple[jax.Array, jax.Array]:
        pass

    @abstractmethod
    def frequency_sweep(self,
            excitation: AbstractExcitation,
            sweep_direction: const.SweepDirection,
        ) -> tuple[jax.Array, jax.Array]:
        pass



# results.py
import matplotlib.pyplot as plt
from dataclasses import dataclass
import jax
import jax.numpy as jnp

from .oscillator.abstract_oscillator import AbstractOscillator
from .solver.abstract_solver import AbstractSolver
from .constants import SweepDirection
from . import constants as const

@jax.tree_util.register_pytree_node_class
@dataclass
class FrequencySweepResult():
    def __init__(self, model: AbstractOscillator, sweep_direction: SweepDirection,
                 driving_frequencies, driving_amplitudes,
                 steady_state_displacement_amplitude, steady_state_velocity_amplitude,
                 total_steady_state_displacement_amplitude, total_steady_state_velocity_amplitude,
                 solver: AbstractSolver):
        
        self.driving_frequencies = driving_frequencies # Shape: (n_driving_frequencies,)
        self.driving_amplitudes = driving_amplitudes # Shape: (n_driving_amplitudes,)
        self.steady_state_displacement_amplitude = steady_state_displacement_amplitude # Shape: (n_driving_frequencies * n_driving_amplitudes, n_modes)
        self.steady_state_velocity_amplitude = steady_state_velocity_amplitude # Shape: (n_driving_frequencies * n_driving_amplitudes, n_modes)
        self.total_steady_state_displacement_amplitude = total_steady_state_displacement_amplitude # Shape: (n_driving_frequencies * n_driving_amplitudes,)
        self.total_steady_state_velocity_amplitude = total_steady_state_velocity_amplitude # Shape: (n_driving_frequencies * n_driving_amplitudes,)
        self.sweep_direction = sweep_direction # SweepDirection Enum

    # PyTree protocol
    def tree_flatten(self):
        children = (
            self.driving_frequencies,
            self.driving_amplitudes,
            self.steady_state_displacement_amplitude,
            self.steady_state_velocity_amplitude,
            self.total_steady_state_displacement_amplitude,
            self.total_steady_state_velocity_amplitude,
        )
        aux = {
            "model": self.model,
            "sweep_direction": self.sweep_direction,
            "solver": self.solver,
        }
        return children, aux

    @classmethod
    def tree_unflatten(cls, aux, children):
        (
            driving_frequencies,
            driving_amplitudes,
            steady_state_displacement_amplitude,
            steady_state_velocity_amplitude,
            total_steady_state_displacement_amplitude,
            total_steady_state_velocity_amplitude,
        ) = children
        return cls(
            model=aux["model"],
            sweep_direction=aux["sweep_direction"],
            driving_frequencies=driving_frequencies,
            driving_amplitudes=driving_amplitudes,
            steady_state_displacement_amplitude=steady_state_displacement_amplitude,
            steady_state_velocity_amplitude=steady_state_velocity_amplitude,
            total_steady_state_displacement_amplitude=total_steady_state_displacement_amplitude,
            total_steady_state_velocity_amplitude=total_steady_state_velocity_amplitude,
            solver=aux["solver"],
        )

    def plot(self):
        raise NotImplementedError("Plotting not implemented for FrequencySweepResult")


import jax
import jax.numpy as jnp
import diffrax
from equinox import filter_jit, filter_checkpoint
import optimistix as optx
import lineax as lx
import numpy as np

from .abstract_solver import AbstractSolver
from ..oscillator.abstract_oscillator import AbstractOscillator
from ..multistart.abstract_multistart import AbstractMultistart
from ..multistart.linear_response_multistart import LinearResponseMultistart
from .. import constants as const 
from .utils.polynomials import LagrangeBasis

class CollocationSolver(AbstractSolver):
    def __init__(self,  max_iterations: int = 20, N_elements: int = 10, m_collocation_points: int = 2, multistart: AbstractMultistart = LinearResponseMultistart(),
                 rtol: float = 1e-4, atol: float = 1e-7, n_time_steps: int = None, max_steps: int = 4096, 
                 verbose: bool = False, throw: bool = False):

        self.n_time_steps = n_time_steps
        self.max_steps = max_steps
        self.max_iterations = max_iterations
        self.N_elements = N_elements
        self.m_collocation_points = m_collocation_points
        self.order_approx = 5
        self.multistart = multistart
        self.rtol = rtol
        self.atol = atol
        self.verbose = verbose
        self.throw = throw

        self.model: AbstractOscillator = None

        self.multistart.verbose = self.verbose

        self.t0 = 0.0
        self.t1 = 1.0
        
        self.ts = jnp.linspace(self.t0, self.t1, self.N_elements * (self.m_collocation_points + 1))  # (N_elements*(m_collocation_points+1),)
        self.ts_segments = self.ts.reshape(self.N_elements, self.m_collocation_points + 1)  # (N_elements, m_collocation_points+1)


        self.he = (self.t1 - self.t0) / self.N_elements 

        self.legendre_nodes, self.legendre_weights = np.polynomial.legendre.leggauss(self.m_collocation_points)  # nodes and weights on [-1,1]
        self.local_collocation_points = 0.5 * (self.legendre_nodes + 1.0)
        self.lagrange_basis = LagrangeBasis(self.m_collocation_points, self.local_collocation_points)

        self.f = jax.vmap(self._rhs, in_axes=(0, 0, None))  # (K+1, n_modes*2)

    def time_response(self,
                 drive_freq: jax.Array,  
                 drive_amp: jax.Array, 
                 init_disp: jax.Array,  
                 init_vel: jax.Array
                ):


        y0_guess = jnp.concatenate([jnp.atleast_1d(init_disp), jnp.atleast_1d(init_vel)], axis=-1)
        Y0 = self._calc_Y0((drive_amp, drive_freq), y0_guess)
        y0, y_max = self._calc_periodic_solution((drive_amp, drive_freq), Y0)

        ts = jnp.linspace(self.t0, self.t1, self.n_time_steps)

        def _solve_one_period(y0):               
            sol = diffrax.diffeqsolve(
                terms=diffrax.ODETerm(self._rhs),
                solver=diffrax.Kvaerno5(),
                t0=self.t0, t1=self.t1, dt0=None, max_steps=self.max_steps,
                y0=y0,
                saveat=diffrax.SaveAt(ts=ts),
                throw=self.throw,
                progress_meter=diffrax.NoProgressMeter(),
                stepsize_controller=diffrax.PIDController(rtol=self.rtol, atol=self.atol),
                args=(drive_amp, drive_freq),
            )
            return sol.ts, sol.ys

        _, ys = jax.lax.cond(
            jnp.isnan(y0).any(),
            lambda: (jnp.zeros((self.n_time_steps,)),
                     jnp.zeros((self.n_time_steps, self.model.n_modes * 2))),
            lambda: _solve_one_period(y0)
        )

        ts = ts * 1 / (drive_freq / (2 * jnp.pi))

        xs = ys[:, :self.model.n_modes]
        vs = ys[:, self.model.n_modes:] / drive_freq

        ys = jnp.concatenate([xs, vs], axis=-1)

        return ts, ys
    
    def frequency_sweep(self,
             f_omega: jax.Array,
             f_amp: jax.Array, 
             sweep_direction: const.SweepDirection,
            ):

        f_omega_mesh, f_amp_mesh, x0_mesh, v0_mesh = self.multistart.generate_simulation_grid(
            self.model, f_omega.flatten(), f_amp.flatten()
        )

        f_omega_mesh_flat = f_omega_mesh.ravel()
        f_amp_mesh_flat  = f_amp_mesh.ravel()
        y0_guess = jnp.stack([x0_mesh, v0_mesh], axis=-1).reshape(-1, self.model.n_states) 

        Y0 = jax.vmap(self._calc_Y0, in_axes=((0,0), 0))(
            (f_amp_mesh_flat, f_omega_mesh_flat), y0_guess
        )  # Y0: (n_sim, N_elements*(m_collocation_points+1), n_states)

        y0, y_max = jax.vmap(self._calc_periodic_solution, in_axes=((0,0), 0))(
            (f_amp_mesh_flat, f_omega_mesh_flat), Y0
        )  # y0: (n_sim, n_states)  y_max: (n_sim, n_modes)

        return {
            "y0": y0,
            "y_max": y_max,
        }
    
    @filter_jit
    def _calc_Y0(self,
            args: jax.Array,
            y0_guess: jax.Array):
                
        sol = diffrax.diffeqsolve(
            terms=diffrax.ODETerm(self._rhs), solver=diffrax.Kvaerno5(),
            t0=self.t0, t1=self.t1, dt0=None,
            y0=y0_guess,
            saveat=diffrax.SaveAt(ts=self.ts),
            throw=self.throw,
            max_steps=self.max_steps,
            progress_meter=diffrax.NoProgressMeter(),
            stepsize_controller=diffrax.PIDController(rtol=self.rtol, atol=self.atol),
            args=args,
        )

        def _process_success(sol):
            Y0 = sol.ys
            return Y0

        def _initial_guess_fallback():
            Y0 = jnp.tile(y0_guess, (self.N_elements * (self.m_collocation_points + 1), 1))
            return Y0

        is_successful = sol.result == diffrax.RESULTS.successful
        is_finite = jnp.all(jnp.isfinite(sol.ys))
        fallback = jnp.logical_not(jnp.logical_and(is_successful, is_finite))
        
        Y0 = jax.lax.cond(
            fallback,
            lambda: _initial_guess_fallback(),
            lambda: _process_success(sol)
        )

        return Y0
    
    @filter_jit
    def _calc_periodic_solution(self,
            args: jax.Array,
            Y0: jax.Array):
        
        def _solve_linear_system(Y):
            jacobian = None

            def _condensation_of_parameters_full_system():
                pass

            def _condensation_of_parameters_uncoupled_system():
                pass

        solver = optx.LevenbergMarquardt(rtol=self.rtol, atol=self.atol, linear_solver=lx.NormalCG(rtol=self.rtol, atol=self.atol), norm=optx.max_norm) 
        sol = optx.least_squares(self._residual, solver, args=args, y0=Y0, max_steps=self.max_iterations, throw=self.throw) 

        def _return_fail():
            y0_fail = jnp.full((self.model.n_modes * 2,), jnp.nan)
            y_max_fail = jnp.full((self.model.n_modes,), jnp.nan)
            return y0_fail, y_max_fail

        def _postprocess_periodic_solution(Y):    
            Y_series = Y.reshape(-1, self.model.n_modes * 2)
            y_max = jnp.max(jnp.abs(Y_series[:, 0::2]), axis=0)

            #self._analyze_periodic_solution(Y[0], args)

            y0 = Y[0]
            return y0, y_max
        
        is_successful = sol.result == optx.RESULTS.successful
        is_finite = jnp.all(jnp.isfinite(sol.value))
        y0, y_max = jax.lax.cond(
            jnp.logical_and(is_successful, is_finite),
            lambda: _postprocess_periodic_solution(sol.value),
            lambda: _return_fail()
        )
        return y0, y_max

    @filter_jit
    def _residual(self, Y, args):
        Y_segments = Y.reshape(self.N_elements, self.m_collocation_points + 1, -1)
        
        def _one_segment_rcol(ts_i, Y_i):
            p, dp_dt = self.lagrange_basis.evaluate(Y_i) # (m, n), (m, n)

            ts_col_i = ts_i[0] + self.he * self.local_collocation_points # (m,)

            r_col = dp_dt - self.he * self.f(ts_col_i, p, args) # (m, n)

            return r_col
    
        _batch_segments = jax.vmap(_one_segment_rcol, in_axes=(0, 0))

        R_col  = _batch_segments(self.ts_segments, Y_segments) # (N, m, n)
        R_cont = Y_segments[1:, 0, :] - Y_segments[:-1, -1, :] # (N-1, n)
        R_per  = Y_segments[0, 0, :] - Y_segments[-1, -1, :] # (n,)

        R = jnp.concatenate([R_col.flatten(), R_cont.flatten(), R_per.flatten()])
        return R
    
    def _jacobian_residual(self, Y, args):
        raise NotImplementedError

    def _analyze_periodic_solution(self):
        raise NotImplementedError

    @filter_jit
    def _rhs(self, t, y, args):
        f_amp, f_omega  = args

        dtau_dt = 2.0 * jnp.pi * (self.model.omega_ref / f_omega)
        tau = dtau_dt * t

        dy_dt = self.model.f(tau, y, args) * dtau_dt

        return dy_dt




from __future__ import annotations
from dataclasses import field, dataclass
from abc import ABC, abstractmethod

oscillator = lambda cls: dataclass(eq=False, kw_only=True)(cls)

@oscillator
class AbstractOscillator(ABC):
    def __post_init__(self):
        pass
    
    @abstractmethod
    def to_dtype(self, dtype):
        pass

    @abstractmethod
    def f(self, t, state, args):
        pass

    @abstractmethod
    def f_y(self, t, state, args):
        pass
    


from __future__ import annotations
import jax
import jax.numpy as jnp
from functools import partial
from typing import Optional

from jax import tree_util

from .. import constants as const
from .abstract_oscillator import AbstractOscillator, oscillator

@oscillator
class VanDerPolOscillator(AbstractOscillator):
    mu: jax.Array

    def __post_init__(self):
        jnp.asarray(self.mu)

    def f(self, tau, state, args, omega_ref=1.0, x_ref=1.0):
        q, dq_dtau   = jnp.split(state, 2)
        f_amp, f_omega = [jnp.asarray(v).squeeze(()) for v in args]

        #omega_ref = f_omega[0]

        damping_term = self.mu * (1 - q**2) * dq_dtau
        linear_stiffness_term = q
        forcing_term = f_amp / jnp.cos(f_omega * tau)

        d2q_dtau2 = (
            + damping_term
            - linear_stiffness_term
            + forcing_term
        ) 
        return jnp.concatenate([dq_dtau, d2q_dtau2])

    def f_y(self, tau, state, args):
        pass
    
    @property
    def n_modes(self) -> int:
        return self.mu.shape[0]
    
    @property
    def n_states(self) -> int:
        return self.n_modes * 2

    def t_steady_state(self, driving_frequency: jax.Array, ss_tol: float) -> float:

        #driving_frequency = 1.0
        t_steady_state = 100.0

        return t_steady_state
    
    def to_dtype(self, dtype: jnp.dtype) -> VanDerPolOscillator:
        return VanDerPolOscillator(
            mu=self.mu.astype(dtype)
        )
    



import jax
import jax.numpy as jnp

def get_unique_solutions(
    y0: jax.Array,
    max_branches: int,
    n: int,
    n_freq: int,
    n_amp: int,
    *,
    rtol: float = 1e-5,
    atol: float = 1e-8,
):
    """
    y0: (n_sim, n) with n_sim = n_freq * n_amp * n_ic
    Returns:
      y0_unique:       (max_branches, n, n_freq, n_amp)  # NaN-padded where unused
      branches_count:  (n_freq, n_amp)                   # number of unique branches kept in each cell
      counts_per_slot: (max_branches, n_freq, n_amp)     # how many rows mapped to each kept branch
    """
    assert y0.ndim == 2 and y0.shape[1] == n, "y0 shape must be (n_sim, n)"
    total_cells = n_freq * n_amp
    assert y0.shape[0] % total_cells == 0, "y0 rows must be divisible by n_freq * n_amp"
    n_ic = y0.shape[0] // total_cells

    # Group rows by (freq, amp) -> shape (n_freq, n_amp, n_ic, n)
    y_grp = y0.reshape(n_freq, n_amp, n_ic, n)

    def per_group(y_group: jax.Array):
        """
        y_group: (n_ic, n)
        Select up to max_branches unique rows under allclose(rtol, atol),
        keeping first occurrences. Returns NaN-padded reps and bookkeeping.
        """
        reps   = jnp.zeros((max_branches, n), dtype=y_group.dtype)          # slots
        valid  = jnp.zeros((max_branches,), dtype=bool)
        counts = jnp.zeros((max_branches,), dtype=jnp.int32)
        M      = jnp.int32(0)  # number of reps filled

        def body(i, carry):
            reps, valid, counts, M = carry
            curr = y_group[i]  # (n,)

            # Compare to existing reps (only where valid=True)
            ref  = jnp.maximum(jnp.abs(reps), jnp.abs(curr))                 # (max_branches, n)
            tol  = atol + rtol * ref
            close = jnp.all(jnp.abs(reps - curr) <= tol, axis=1) & valid     # (max_branches,)

            candidates  = jnp.where(close, jnp.arange(max_branches, dtype=jnp.int32), max_branches)
            first_match = jnp.min(candidates)
            matched     = first_match < max_branches
            has_space   = M < max_branches
            create_new  = (~matched) & has_space
            slot        = jnp.where(matched, first_match, M)

            reps  = jax.lax.cond(create_new, lambda r: r.at[slot].set(curr), lambda r: r, reps)
            valid = jax.lax.cond(create_new, lambda v: v.at[slot].set(True),  lambda v: v, valid)
            counts = counts.at[slot].add(1)
            M = jax.lax.select(create_new, M + 1, M)
            return reps, valid, counts, M

        reps, valid, counts, M = jax.lax.fori_loop(0, y_group.shape[0], body,
                                                   (reps, valid, counts, jnp.int32(0)))

        # Pad unused slots with NaN for clarity
        reps = jnp.where(valid[:, None], reps, jnp.array(jnp.nan, reps.dtype))

        return reps, valid, counts, M

    # Vectorize over (freq, amp)
    process_all = jax.jit(
        jax.vmap(
            jax.vmap(per_group, in_axes=0, out_axes=(0, 0, 0, 0)),  # over n_amp
            in_axes=0, out_axes=(0, 0, 0, 0)                        # over n_freq
        )
    )

    reps, valid, counts, M = process_all(y_grp)
    # reps:   (n_freq, n_amp, max_branches, n)
    # valid:  (n_freq, n_amp, max_branches)  (not returned, but could be)
    # counts: (n_freq, n_amp, max_branches)
    # M:      (n_freq, n_amp)

    # Reorder to requested layout
    y0_unique      = jnp.transpose(reps, (2, 3, 0, 1))            # (max_branches, n, n_freq, n_amp)
    counts_per_slot = jnp.transpose(counts, (2, 0, 1))            # (max_branches, n_freq, n_amp)
    branches_count  = M                                            # (n_freq, n_amp)

    return y0_unique, branches_count, counts_per_slot



# Copyright (c) 2026 Raymond Knetemann
# Licensed under the MIT License. See LICENSE file for details.

from __future__ import annotations
import jax
import jax.numpy as jnp
import time

from .oscillator.abstract_oscillator import AbstractOscillator
from .solver.abstract_solver import AbstractSolver
from .solver.time_integration_solver import TimeIntegrationSolver
from . import constants as const

def time_response(
    model: AbstractOscillator,
    driving_frequency: jax.Array, # Shape: (1,)
    driving_amplitude: jax.Array, # Shape: (n_modes,)
    initial_displacement: jax.Array, # Shape: (n_modes,)
    initial_velocity: jax.Array, # Shape: (n_modes,)
    solver: AbstractSolver = TimeIntegrationSolver(),
    precision: const.Precision = const.Precision.DOUBLE,
    **kwargs
) -> tuple:
    """Compute the time response of a dynamical model to a one-tone excitation.

    Args:
        model: The dynamical model to simulate.
        driving_frequency: The driving frequency (shape: (1,)).
        driving_amplitude: The driving amplitude for each mode (shape: (n_modes,)).
        initial_displacement: The initial displacement for each mode (shape: (n_modes,)).
        initial_velocity: The initial velocity for each mode (shape: (n_modes,)).
        solver: The time integration solver to use.
        precision: The numerical precision to use.
        **kwargs: Additional keyword arguments to pass to the solver's time_response method.

    Returns:
        A tuple (ts, xs, vs) where ts is the time array (shape: (n_steps,)),
        xs is the displacement array (shape: (n_steps, n_modes)), and vs is the velocity array (shape: (n_steps, n_modes)).
    """
    
    if precision == const.Precision.DOUBLE:
        jax.config.update("jax_enable_x64", True)
        dtype = jnp.float64
    elif precision == const.Precision.SINGLE:
        jax.config.update("jax_enable_x64", False)
        dtype = jnp.float32
    else:
        raise ValueError(f"Unsupported precision: {precision}")

    # Ensure inputs are jax arrays (handles Python floats, lists, numpy arrays) before dtype enforcement
    driving_frequency = jnp.asarray(driving_frequency, dtype=dtype)
    driving_amplitude = jnp.asarray(driving_amplitude, dtype=dtype)
    initial_displacement = jnp.asarray(initial_displacement, dtype=dtype)
    initial_velocity = jnp.asarray(initial_velocity, dtype=dtype)
    
    if model.n_modes != initial_displacement.size:
        raise ValueError(f"Model has {model.n_modes} modes, but initial displacement has shape {initial_displacement.shape}. It should have shape ({model.n_modes},).")
    if model.n_modes != initial_velocity.size:
        raise ValueError(f"Model has {model.n_modes} modes, but initial velocity has shape {initial_velocity.shape}. It should have shape ({model.n_modes},).")
    
    model = model.to_dtype(dtype)
    solver.model = model

    print("Time response: ", model)
    start_time = time.time()
    ts, ys = solver.time_response(driving_frequency, driving_amplitude, initial_displacement, initial_velocity, **kwargs)
    print("Time response completed in {:.2f} seconds".format(time.time() - start_time))


    xs = ys[:, :model.n_modes]  # Shape: (n_steps, n_modes)
    vs = ys[:, model.n_modes:]  # Shape: (n_steps, n_modes)

    return ts, xs, vs


import jax.numpy as jnp

from .abstract_multistart import AbstractMultistart

class LinearResponseMultistart(AbstractMultistart):
    def __init__(self, init_cond_shape: tuple = (3, 3), linear_response_factor: float = 1.0):
        super().__init__()
        self.init_cond_shape = init_cond_shape
        self.linear_response_factor = linear_response_factor

    def generate_simulation_grid(self, model, f_omegas, f_amps):
        max_force = jnp.max(f_amps, axis=None)
        max_displacement_per_mode = (max_force * model.Q / model.omega_0**2) * self.linear_response_factor

        n_f_omegas = f_omegas.shape[0]
        n_f_amps = f_amps.shape[0]
        n_x0s = self.init_cond_shape[0]
        n_v0s = self.init_cond_shape[1]
        n_modes = model.n_modes

        f_omegas_grid = jnp.tile(f_omegas[:, None], (1, n_modes))
        f_amps_grid = f_amps

        if n_x0s > 1:
            x0s_grid = jnp.outer(jnp.linspace(
                -1.0, 1.0, n_x0s
            ), max_displacement_per_mode)

        else:
            x0s_grid = jnp.zeros((1, n_modes))

        if n_v0s > 1:
            v0s_grid = jnp.outer(jnp.linspace(
                -1.0, 1.0, n_v0s
            ), max_displacement_per_mode)
        else:
            v0s_grid = jnp.zeros((1, n_modes))

        f_omega_idx_mesh, f_amp_idx_mesh, x0_idx_mesh, v0_idx_mesh, mode_idx_mesh = jnp.meshgrid(
            jnp.arange(n_f_omegas),
            jnp.arange(n_f_amps),
            jnp.arange(n_x0s),
            jnp.arange(n_v0s),
            jnp.arange(n_modes),
            indexing="ij",
        )

        f_omegas_mesh = f_omegas_grid[f_omega_idx_mesh, mode_idx_mesh]
        f_amps_mesh = f_amps_grid[f_amp_idx_mesh, mode_idx_mesh]
        x0_mesh = x0s_grid[x0_idx_mesh, mode_idx_mesh]
        v0_mesh = v0s_grid[v0_idx_mesh, mode_idx_mesh]

        n_combinations = n_f_omegas * n_f_amps * n_x0s * n_v0s
        f_omegas_mesh = f_omegas_mesh.reshape(n_combinations, n_modes)
        f_amps_mesh = f_amps_mesh.reshape(n_combinations, n_modes)
        x0_mesh = x0_mesh.reshape(n_combinations, n_modes)
        v0_mesh = v0_mesh.reshape(n_combinations, n_modes)

        shape = (n_f_omegas, n_f_amps, n_x0s, n_v0s, n_modes)  # Update shape to reflect the full grid dimensions

        return (f_omegas_mesh, f_amps_mesh, x0_mesh, v0_mesh, shape)
    
    def to_dtype(self, dtype: jnp.dtype):
        return LinearResponseMultistart(
            init_cond_shape=self.init_cond_shape,
            linear_response_factor=self.linear_response_factor
        )


import numpy as np
import poscidyn
import time

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

def F_max (eta, omega_0, Q, gamma):
    return np.sqrt(4 * omega_0**6 / (3 * gamma * Q**2) * (eta + 1 / (2*Q**2)) * (1 + eta + 1 / (4 * Q **2)))

# 1 mode:
Q, omega_0, alpha, gamma = np.array([100.0]), np.array([1.00]), np.zeros((1,1,1)), np.zeros((1,1,1,1))
gamma[0,0,0,0] = 2.55
modal_forces = np.array([1.0])

# 2 modes:
Q, omega_0, alpha, gamma = np.array([6.357357, 36.742565]), np.array([0.9946662, 2.1275177]), np.zeros((2,2,2)), np.zeros((2,2,2,2))
gamma[0,0,0,0] = 1.9451260e+01
gamma[0,0,1,1] = 1.3927963e+02
gamma[1,1,1,1] = 2.6890811e+02
gamma[1,0,0,1] = 1.4860869e+02

# 3 modes:
# Q, omega_0, alpha, gamma = np.array([100.0, 100.0, 100.0]), np.array([1.00, 1.73, 2.59]), np.zeros((3,3,3)), np.zeros((3,3,3,3))
# gamma[0,0,0,0] = 2.55
# gamma[0,0,1,1] = 8.61
# gamma[0,0,0,2] = 0.942
# gamma[0,0,1,2] = -0.143
# gamma[0,0,2,2] = 9.88
# gamma[1,1,1,1] = 18.7
# gamma[1,0,0,1] = 8.57
# gamma[1,0,0,2] = -0.000283
# gamma[1,0,1,2] = -11.8
# gamma[1,0,2,2] = 0.000849
# gamma[1,1,1,2] = -0.00272
# gamma[1,1,2,2] = 45.2
# gamma[2,2,2,2] = -0.00144

F_max_value = F_max(0.20, omega_0[0], Q[0], gamma[0,0,0,0])

print(f"Calculated F_max: {F_max_value:.4f}")

driving_frequency = np.linspace(0.6322645545005798, 2.2314629554748535, 400)
driving_amplitude = np.linspace(0.1, 1.0, 10) * 1
modal_forces = np.array([0.01101661, 0.00740971])

MODEL = poscidyn.NonlinearOscillator(Q=Q, alpha=alpha, gamma=gamma, omega_0=omega_0)
EXCITOR = poscidyn.OneToneExcitation(driving_frequency, driving_amplitude, modal_forces)
MULTISTART = poscidyn.LinearResponseMultistart(init_cond_shape=(3, 3), linear_response_factor=1.0)
SOLVER = poscidyn.TimeIntegrationSolver(max_steps=4096*5, n_time_steps=50, verbose=True, throw=False, rtol=1e-4, atol=1e-7)
SWEEPER = poscidyn.NearestNeighbourSweep(sweep_direction=[poscidyn.Forward(), poscidyn.Backward()])
PRECISION = poscidyn.Precision.SINGLE

def _extract_gamma_diagonal(gamma: np.ndarray):
    arr = np.asarray(gamma)
    if arr.ndim < 4:
        return arr.ravel().tolist()[: arr.size]

    diag = []
    max_modes = min(2, arr.shape[0], arr.shape[1], arr.shape[2], arr.shape[3])
    for i in range(max_modes):
        diag.append(float(arr[i, i, i, i]))
    return diag


def _extract_gamma_entries(gamma: np.ndarray, max_entries: int = 3):
    arr = np.asarray(gamma)
    non_zero = np.argwhere(arr != 0)
    entries = []
    for idx in non_zero[:max_entries]:
        tuple_idx = tuple(idx.tolist())
        entries.append((tuple_idx, float(arr[tuple_idx])))
    return entries, int(non_zero.shape[0])


def _extract_alpha_entries(alpha: np.ndarray, max_entries: int = 3):
    arr = np.asarray(alpha)
    non_zero = np.argwhere(arr != 0)
    entries = []
    for idx in non_zero[:max_entries]:
        tuple_idx = tuple(idx.tolist())
        entries.append((tuple_idx, float(arr[tuple_idx])))
    return entries, int(non_zero.shape[0])


def _format_param_text(Q: np.ndarray, omega_0: np.ndarray, alpha: np.ndarray, gamma: np.ndarray, modal_forces: np.ndarray) -> str:
    q_vals = np.asarray(Q).ravel()
    omega_vals = np.asarray(omega_0).ravel()
    alpha_entries, alpha_total = _extract_alpha_entries(alpha)
    gamma_entries, gamma_total = _extract_gamma_entries(gamma)

    parts = []
    if q_vals.size:
        parts.append(f"Q=[{', '.join(f'{val:.2f}' for val in q_vals[:2])}]")
    if omega_vals.size:
        parts.append(f"omega0=[{', '.join(f'{val:.2f}' for val in omega_vals[:2])}]")
    if alpha_entries:
        formatted_alpha = ", ".join(f"{idx}={val:.2e}" for idx, val in alpha_entries)
        if alpha_total > len(alpha_entries):
            formatted_alpha += ", ..."
        parts.append(f"alpha={formatted_alpha}")
    if gamma_entries:
        formatted_gamma = ", ".join(f"{idx}={val:.2e}" for idx, val in gamma_entries)
        if gamma_total > len(gamma_entries):
            formatted_gamma += ", ..."
        parts.append(f"gamma={formatted_gamma}")
    if modal_forces.size:
        parts.append(f"modal_forces=[{', '.join(f'{val:.2f}' for val in modal_forces[:2])}]")
    return "\n".join(parts)


def plot_sweep(ax, drive_freqs, drive_amps, sweeped_solutions, param_text: str) -> None:
    forward = sweeped_solutions.get("forward")
    backward = sweeped_solutions.get("backward")

    if forward is None and backward is None:
        raise ValueError("No sweeped solutions to plot.")

    ref_idx = np.argmin(np.abs(drive_freqs - 0.9 * omega_0[0]))

    omega_ref = drive_freqs[ref_idx]
    x_ref_forward = forward[ref_idx, :]
    x_ref_backward = backward[ref_idx, :]
    print(f"Reference frequency: {omega_ref}")
    print(f"Reference displacement forward: {x_ref_forward}, backward: {x_ref_backward}")

    norm_alpha_forward = (x_ref_forward / omega_ref**2)
    norm_alpha_backward = (x_ref_backward / omega_ref**2)
    norm_gamma_forward = (x_ref_forward**2 / omega_ref**2)
    norm_gamma_backward = (x_ref_backward**2 / omega_ref**2)

    alpha_ndim_forward  = norm_alpha_forward[:, None, None, None] * alpha[None, :, :, :] 
    alpha_ndim_backward = norm_alpha_backward[:, None, None, None] * alpha[None, :, :, :]

    gamma_ndim_forward  = norm_gamma_forward[:, None, None, None, None] * gamma[None, :, :, :, :] 
    gamma_ndim_backward = norm_gamma_backward[:, None, None, None, None] * gamma[None, :, :, :, :]  

    forward = forward / x_ref_forward
    backward = backward / x_ref_backward

    drive_freqs = np.asarray(drive_freqs) / omega_ref
    drive_amps = np.asarray(drive_amps)
    colors = plt.cm.viridis(np.linspace(0, 1, drive_amps.size))

    for idx, (amp, color) in enumerate(zip(drive_amps, colors)):
        if forward is not None:
            ax.plot(
                drive_freqs,
                forward[:, idx],
                color=color,
                linestyle="-",
                linewidth=1.3,
            )
        if backward is not None:
            ax.plot(
                drive_freqs,
                backward[:, idx],
                color=color,
                linestyle="--",
                linewidth=1.1,
            )

    ax.set_title(
        f"Frequency sweep"
    )
    ax.set_xlabel("Drive frequency")
    ax.set_ylabel("Max displacement")
    ax.grid(alpha=0.25)

    if param_text:
        ax.text(
            0.02,
            0.98,
            param_text,
            transform=ax.transAxes,
            fontsize=8,
            va="top",
            ha="left",
            bbox=dict(facecolor="white", edgecolor="none", alpha=0.65),
        )

    amp_handles = [
        Line2D([0], [0], color=color, lw=1.8) for color in colors
    ]
    amp_labels = [f"F={amp:.3f}" for amp in drive_amps]
    legend1 = ax.legend(
        amp_handles,
        amp_labels,
        title="Drive amplitude",
        loc="upper right",
        bbox_to_anchor=(1.0, 1.0),
        fontsize=7,
        frameon=False,
        ncol=2,
    )

    if forward is not None and backward is not None:
        style_handles = [
            Line2D([0], [0], color="k", linestyle="-", lw=1.5),
            Line2D([0], [0], color="k", linestyle="--", lw=1.5),
        ]
        style_labels = ["Forward sweep", "Backward sweep"]
        legend2 = ax.legend(
            style_handles,
            style_labels,
            loc="lower right",
            fontsize=7,
            frameon=False,
        )
        ax.add_artist(legend1)
    else:
        ax.add_artist(legend1)

start_time = time.time()

frequency_sweep = poscidyn.frequency_sweep(
    model = MODEL,
    sweeper=SWEEPER,
    excitor=EXCITOR,
    solver = SOLVER,
    precision = PRECISION,
    multistarter=MULTISTART,
) #n_freq, n_amp, n_init_disp, n_init_vel

end_time = time.time()
print(f"Frequency sweep completed in {end_time - start_time:.2f} seconds.")
print(
    f"Successful periodic solutions: {frequency_sweep.n_successful}/{frequency_sweep.n_total} "
    f"({frequency_sweep.success_rate:.1%})"
)


fig, ax = plt.subplots(figsize=(10, 6))
plot_sweep(
    ax=ax,
    drive_freqs=EXCITOR.drive_frequencies,
    drive_amps=EXCITOR.drive_amplitudes,
    sweeped_solutions=frequency_sweep.sweeped_periodic_solutions,
    param_text=_format_param_text(Q, omega_0, alpha, gamma, EXCITOR.modal_forces),
)
plt.tight_layout()
plt.show()


from jax import tree_util

class FrequencySweepResult:
    def __init__(self, f_omegas, f_amps, modal_forces, Q, omega_0, alpha, gamma, periodic_solutions, sweeped_periodic_solutions, n_successful, n_total, success_rate):
        self.f_omegas = f_omegas
        self.f_amps = f_amps
        self.modal_forces = modal_forces
        self.Q = Q
        self.omega_0 = omega_0
        self.alpha = alpha
        self.gamma = gamma
        self.periodic_solutions = periodic_solutions
        self.sweeped_periodic_solutions = sweeped_periodic_solutions
        self.n_successful = n_successful
        self.n_total = n_total
        self.success_rate = success_rate

# Register FrequencySweepResult as a pytree
def _tree_flatten(obj):
    leaves = (
        obj.f_omegas,
        obj.f_amps,
        obj.modal_forces,
        obj.Q,
        obj.omega_0,
        obj.alpha,
        obj.gamma,
        obj.periodic_solutions,
        obj.sweeped_periodic_solutions,
        obj.n_successful,
        obj.n_total,
        obj.success_rate,
    )
    return leaves, None

def _tree_unflatten(aux_data, leaves):
    (
        f_omegas,
        f_amps,
        modal_forces,
        Q,
        omega_0,
        alpha,
        gamma,
        periodic_solutions,
        sweeped_periodic_solutions,
        n_successful,
        n_total,
        success_rate,
    ) = leaves
    return FrequencySweepResult(
        f_omegas,
        f_amps,
        modal_forces,
        Q,
        omega_0,
        alpha,
        gamma,
        periodic_solutions,
        sweeped_periodic_solutions,
        n_successful,
        n_total,
        success_rate,
    )

tree_util.register_pytree_node(
    FrequencySweepResult, _tree_flatten, _tree_unflatten
)



import jax
import jax.numpy as jnp

from ... import constants as const
from ...oscillator.abstract_oscillator import AbstractOscillator

# TO DO: Randomize initial conditions within the grid cells

def gen_coarse_grid_1(model: AbstractOscillator,
                    drive_freq: jax.Array,
                    drive_amp: jax.Array,
                    ):
    
    coarse_drive_freq = jnp.linspace(
        jnp.min(drive_freq), jnp.max(drive_freq), const.N_COARSE_DRIVING_FREQUENCIES
    ) # (N_COARSE_DRIVING_FREQUENCIES,)

    coarse_drive_amp = jnp.linspace(
        jnp.min(drive_amp), jnp.max(drive_amp), const.N_COARSE_DRIVING_AMPLITUDES
    ) # (N_COARSE_DRIVING_AMPLITUDES,)

    # For each combination of coarse driving frequency and amplitude, estimate the linear response amplitude
    denom = jnp.sqrt((model.omega_0**2 - coarse_drive_freq**2)**2 + (model.omega_0 * coarse_drive_freq / model.Q)**2)
    linear_response_amplitude = coarse_drive_amp[:, None] / denom[None, :]

    jax.debug.print("Max linear response amplitude: {}", jnp.max(linear_response_amplitude))

    factor = const.N_COARSE_INITIAL_CONDITIONS_OFFSET_FACTOR
    init_disp_grid = jnp.linspace(-factor*linear_response_amplitude, factor*linear_response_amplitude, const.N_COARSE_INITIAL_DISPLACEMENTS, axis=-1)
    init_vel_grid = jnp.linspace(-factor*linear_response_amplitude, factor*linear_response_amplitude, const.N_COARSE_INITIAL_VELOCITIES, axis=-1)

    F = const.N_COARSE_DRIVING_FREQUENCIES
    A = const.N_COARSE_DRIVING_AMPLITUDES
    ND = const.N_COARSE_INITIAL_DISPLACEMENTS
    NV = const.N_COARSE_INITIAL_VELOCITIES

    init_disp_FA = jnp.transpose(init_disp_grid, (1, 0, 2))  # (F, A, ND)
    init_vel_FA  = jnp.transpose(init_vel_grid,  (1, 0, 2))  # (F, A, NV)

    # 2) Broadcast everything to (F, A, ND, NV)
    coarse_drive_freq_mesh = jnp.broadcast_to(
        coarse_drive_freq[:, None, None, None], (F, A, ND, NV)
    )
    coarse_drive_amp_mesh = jnp.broadcast_to(
        coarse_drive_amp[None, :, None, None], (F, A, ND, NV)
    )
    coarse_init_disp_mesh = jnp.broadcast_to(
        init_disp_FA[..., None], (F, A, ND, NV)
    )
    coarse_init_vel_mesh = jnp.broadcast_to(
        init_vel_FA[:, :, None, :], (F, A, ND, NV)
    )

    return (coarse_drive_freq_mesh, coarse_drive_amp_mesh, coarse_init_disp_mesh, coarse_init_vel_mesh)

def gen_grid_2(model: AbstractOscillator,
                    drive_freq: jax.Array,
                    drive_amp: jax.Array,
                    ):
    
    max_abs_displacement = float((jnp.max(drive_amp) * jnp.abs(model.Q)).item()) * const.LINEAR_RESPONSE_FACTOR

    init_disp_grid = jnp.linspace(
        0.0, max_abs_displacement, const.N_COARSE_INITIAL_DISPLACEMENTS
    ) # (N_COARSE_INITIAL_DISPLACEMENTS,)
    
    init_vel_grid = jnp.linspace(
        -max_abs_displacement, max_abs_displacement, const.N_COARSE_INITIAL_VELOCITIES
    ) # (N_COARSE_INITIAL_VELOCITIES,)

    drive_freq_mesh, drive_amp_mesh, init_disp_mesh, init_vel_mesh = jnp.meshgrid(
        drive_freq, drive_amp, init_disp_grid, init_vel_grid, indexing="ij"
    ) # (N_COARSE_DRIVING_FREQUENCIES, N_COARSE_DRIVING_AMPLITUDES, N_COARSE_INITIAL_DISPLACEMENTS, N_COARSE_INITIAL_VELOCITIES)

    return (drive_freq_mesh, drive_amp_mesh, init_disp_mesh, init_vel_mesh)


import numpy as np
import poscidyn
import time
import h5py  # <-- NEW

import matplotlib.pyplot as plt
from matplotlib.lines import Line2D


# =========================
# NEW: load original A from file
# =========================
A_NPZ_PATH = "/home/raymo/Projects/nlsid/worst_gamma_case_A_only.npz"  # <-- set your npz path
HDF5_DATASET_PATH = "/home/raymo/Projects/nlsid/datasets/2026012301_string_2DOF/2026012301_string_2DOF_augmented_split.hdf5"  # <-- set your dataset file
SPLIT = "test"          # <-- set train/val/test
HDF5_INDEX = None       # <-- optional: override; if None we'll use worst_index from the npz

# If you want to plot the dataset A as-is (already normalized), keep True.
# If you want to undo the normalization used in plot_sweep() for the simulation (divide by x_ref),
# set this False and we'll normalize the dataset A the same way before plotting.
PLOT_DATASET_A_AS_STORED = True


def F_max(eta, omega_0, Q, gamma):
    return np.sqrt(
        4 * omega_0**6
        / (3 * gamma * Q**2)
        * (eta + 1 / (2 * Q**2))
        * (1 + eta + 1 / (4 * Q**2))
    )


# 1 mode:
Q, omega_0, alpha, gamma = np.array([100.0]), np.array([1.00]), np.zeros((1, 1, 1)), np.zeros((1, 1, 1, 1))
gamma[0, 0, 0, 0] = 2.55
modal_forces = np.array([1.0])

# 2 modes:
Q, omega_0, alpha, gamma = (
    np.array([6.357357, 36.742565]),
    np.array([0.9946662, 2.1275177]),
    np.zeros((2, 2, 2)),
    np.zeros((2, 2, 2, 2)),
)
gamma[0, 0, 0, 0] = 1.9451260e01
gamma[0, 0, 1, 1] = 1.3927963e02
gamma[1, 1, 1, 1] = 2.6890811e02
gamma[1, 0, 0, 1] = 1.4860869e02

F_max_value = F_max(0.20, omega_0[0], Q[0], gamma[0, 0, 0, 0])
print(f"Calculated F_max: {F_max_value:.4f}")


# You can keep these as you had them:
driving_frequency = np.linspace(0.6322645545005798, 2.2314629554748535, 400)
driving_amplitude = np.linspace(0.1, 1.0, 10) * 1
modal_forces = np.array([0.01101661, 0.00740971])

MODEL = poscidyn.NonlinearOscillator(Q=Q, alpha=alpha, gamma=gamma, omega_0=omega_0)
EXCITOR = poscidyn.OneToneExcitation(driving_frequency, driving_amplitude, modal_forces)
MULTISTART = poscidyn.LinearResponseMultistart(init_cond_shape=(3, 3), linear_response_factor=1.0)
SOLVER = poscidyn.TimeIntegrationSolver(
    max_steps=4096 * 5, n_time_steps=50, verbose=True, throw=False, rtol=1e-4, atol=1e-7
)
SWEEPER = poscidyn.NearestNeighbourSweep(sweep_direction=[poscidyn.Forward(), poscidyn.Backward()])
PRECISION = poscidyn.Precision.SINGLE


def _extract_gamma_entries(gamma: np.ndarray, max_entries: int = 3):
    arr = np.asarray(gamma)
    non_zero = np.argwhere(arr != 0)
    entries = []
    for idx in non_zero[:max_entries]:
        tuple_idx = tuple(idx.tolist())
        entries.append((tuple_idx, float(arr[tuple_idx])))
    return entries, int(non_zero.shape[0])


def _extract_alpha_entries(alpha: np.ndarray, max_entries: int = 3):
    arr = np.asarray(alpha)
    non_zero = np.argwhere(arr != 0)
    entries = []
    for idx in non_zero[:max_entries]:
        tuple_idx = tuple(idx.tolist())
        entries.append((tuple_idx, float(arr[tuple_idx])))
    return entries, int(non_zero.shape[0])


def _format_param_text(Q: np.ndarray, omega_0: np.ndarray, alpha: np.ndarray, gamma: np.ndarray, modal_forces: np.ndarray) -> str:
    q_vals = np.asarray(Q).ravel()
    omega_vals = np.asarray(omega_0).ravel()
    alpha_entries, alpha_total = _extract_alpha_entries(alpha)
    gamma_entries, gamma_total = _extract_gamma_entries(gamma)

    parts = []
    if q_vals.size:
        parts.append(f"Q=[{', '.join(f'{val:.2f}' for val in q_vals[:2])}]")
    if omega_vals.size:
        parts.append(f"omega0=[{', '.join(f'{val:.2f}' for val in omega_vals[:2])}]")
    if alpha_entries:
        formatted_alpha = ", ".join(f"{idx}={val:.2e}" for idx, val in alpha_entries)
        if alpha_total > len(alpha_entries):
            formatted_alpha += ", ..."
        parts.append(f"alpha={formatted_alpha}")
    if gamma_entries:
        formatted_gamma = ", ".join(f"{idx}={val:.2e}" for idx, val in gamma_entries)
        if gamma_total > len(gamma_entries):
            formatted_gamma += ", ..."
        parts.append(f"gamma={formatted_gamma}")
    if modal_forces.size:
        parts.append(f"modal_forces=[{', '.join(f'{val:.2f}' for val in modal_forces[:2])}]")
    return "\n".join(parts)


# =========================
# NEW: load the stored dataset curve + its omega range
# =========================
def load_dataset_curve():
    idx = HDF5_INDEX
    if idx is None:
        # Try to pull index from the npz you saved earlier
        d = np.load(A_NPZ_PATH)
        if "worst_index" in d:
            idx = int(d["worst_index"])
        else:
            raise ValueError("HDF5_INDEX is None and NPZ file does not contain 'worst_index'.")

    with h5py.File(HDF5_DATASET_PATH, "r") as f:
        g = f[SPLIT]
        A_ds = np.asarray(g["A"][idx, :], dtype=np.float64) * (50.0 / 2.347)* 0.9939844
        fmin = float(np.asarray(g["f_omega_min"][idx]))
        fmax = float(np.asarray(g["f_omega_max"][idx]))
        omega_axis = np.linspace(fmin, fmax, A_ds.shape[0], dtype=np.float64)

    return idx, omega_axis, A_ds, fmin, fmax


# =========================
# MODIFIED: plot_sweep now also plots dataset A
# =========================
def plot_sweep(ax, drive_freqs, drive_amps, sweeped_solutions, param_text: str) -> None:
    forward = sweeped_solutions.get("forward")
    backward = sweeped_solutions.get("backward")

    if forward is None and backward is None:
        raise ValueError("No sweeped solutions to plot.")

    # --- reference normalization for the simulation (your existing logic)
    ref_idx = np.argmin(np.abs(drive_freqs - 0.9 * omega_0[0]))
    omega_ref = drive_freqs[ref_idx]

    x_ref_forward = forward[ref_idx, :]
    x_ref_backward = backward[ref_idx, :]

    print(f"Reference frequency: {omega_ref}")
    print(f"Reference displacement forward: {x_ref_forward}, backward: {x_ref_backward}")

    forward = forward / x_ref_forward
    backward = backward / x_ref_backward

    drive_freqs_norm = np.asarray(drive_freqs, dtype=np.float64) / omega_ref
    drive_amps = np.asarray(drive_amps)
    colors = plt.cm.viridis(np.linspace(0, 1, drive_amps.size))

    # --- simulation curves
    for idx, (amp, color) in enumerate(zip(drive_amps, colors)):
        if forward is not None:
            ax.plot(drive_freqs_norm, forward[:, idx], color=color, linestyle="-", linewidth=1.3)
        if backward is not None:
            ax.plot(drive_freqs_norm, backward[:, idx], color=color, linestyle="--", linewidth=1.1)

    # --- NEW: overlay dataset A (as saved)
    ds_idx, omega_ds, A_ds, fmin_ds, fmax_ds = load_dataset_curve()

    # dataset frequency axis normalized the same way as simulation:
    omega_ds_norm = omega_ds / omega_ref

    if PLOT_DATASET_A_AS_STORED:
        A_ds_plot = A_ds
    else:
        # normalize dataset A with the same x_ref strategy:
        # (use the dataset’s own omega axis to find nearest 0.9*omega0[0])
        target = 0.9 * float(omega_0[0])
        ref_idx_ds = int(np.argmin(np.abs(omega_ds - target)))
        x_ref_ds = float(A_ds[ref_idx_ds])
        A_ds_plot = A_ds / x_ref_ds

    ax.plot(
        omega_ds_norm,
        A_ds_plot,
        color="black",
        linewidth=2.2,
        alpha=0.9,
        label=f"Dataset A (idx={ds_idx})",
        zorder=6,
    )

    # also mark the dataset reference point (0.9*omega0[0]) on the dataset curve
    target = 0.9 * float(omega_0[0])
    ref_idx_ds = int(np.argmin(np.abs(omega_ds - target)))
    ax.scatter(
        [omega_ds_norm[ref_idx_ds]],
        [A_ds_plot[ref_idx_ds]],
        color="red",
        s=55,
        zorder=7,
        label="0.9·ω₀[0] ref (dataset)",
    )

    ax.set_title("Frequency sweep (simulation) + stored dataset A overlay")
    ax.set_xlabel("Drive frequency (normalized by ω_ref)")
    ax.set_ylabel("Max displacement (simulation normalized by x_ref)")
    ax.grid(alpha=0.25)

    if param_text:
        ax.text(
            0.02,
            0.98,
            param_text,
            transform=ax.transAxes,
            fontsize=8,
            va="top",
            ha="left",
            bbox=dict(facecolor="white", edgecolor="none", alpha=0.65),
        )

    # Legends
    amp_handles = [Line2D([0], [0], color=color, lw=1.8) for color in colors]
    amp_labels = [f"F={amp:.3f}" for amp in drive_amps]
    legend1 = ax.legend(
        amp_handles,
        amp_labels,
        title="Drive amplitude",
        loc="upper right",
        bbox_to_anchor=(1.0, 1.0),
        fontsize=7,
        frameon=False,
        ncol=2,
    )

    style_handles = [
        Line2D([0], [0], color="k", linestyle="-", lw=1.5),
        Line2D([0], [0], color="k", linestyle="--", lw=1.5),
        Line2D([0], [0], color="black", linestyle="-", lw=2.2),
        Line2D([0], [0], color="red", marker="o", linestyle="None", markersize=6),
    ]
    style_labels = ["Forward sweep", "Backward sweep", "Dataset A", "Dataset ref point"]

    legend2 = ax.legend(
        style_handles,
        style_labels,
        loc="lower right",
        fontsize=7,
        frameon=False,
    )
    ax.add_artist(legend1)


start_time = time.time()

frequency_sweep = poscidyn.frequency_sweep(
    model=MODEL,
    sweeper=SWEEPER,
    excitor=EXCITOR,
    solver=SOLVER,
    precision=PRECISION,
    multistarter=MULTISTART,
)  # n_freq, n_amp, n_init_disp, n_init_vel

end_time = time.time()
print(f"Frequency sweep completed in {end_time - start_time:.2f} seconds.")
print(
    f"Successful periodic solutions: {frequency_sweep.n_successful}/{frequency_sweep.n_total} "
    f"({frequency_sweep.success_rate:.1%})"
)

fig, ax = plt.subplots(figsize=(10, 6))
plot_sweep(
    ax=ax,
    drive_freqs=EXCITOR.drive_frequencies,
    drive_amps=EXCITOR.drive_amplitudes,
    sweeped_solutions=frequency_sweep.sweeped_periodic_solutions,
    param_text=_format_param_text(Q, omega_0, alpha, gamma, EXCITOR.modal_forces),
)
plt.tight_layout()
plt.show()



import jax
import jax.numpy as jnp
import diffrax
from equinox import filter_jit
import optimistix as optx
import lineax as lx

from .abstract_solver import AbstractSolver
from ..oscillator.abstract_oscillator import AbstractOscillator
from ..multistart.abstract_multistart import AbstractMultistart
from ..multistart.linear_response_multistart import LinearResponseMultistart
from .. import constants as const 

class MultipleShootingSolver(AbstractSolver):
    def __init__(self,  max_shooting_iterations: int = 20, m_segments: int = 20, multistart: AbstractMultistart = LinearResponseMultistart(),
                 rtol: float = 1e-4, atol: float = 1e-7, n_time_steps: int = None, max_steps: int = 4096, verbose: bool = False):

        self.n_time_steps = n_time_steps
        self.max_steps = max_steps
        self.max_shooting_iterations = max_shooting_iterations
        self.m_segments = m_segments
        self.multistart = multistart
        self.rtol = rtol
        self.atol = atol
        self.verbose = verbose

        self.model: AbstractOscillator = None

        self.multistart.verbose = self.verbose

        self.T = 2.0 * jnp.pi
        self.t0 = 0.0
        self.t1 = self.T
   
    def time_response(self,
                 drive_freq: jax.Array,  
                 drive_amp: jax.Array, 
                 init_disp: jax.Array,  
                 init_vel: jax.Array
                ):

        y0_guess = jnp.array([init_disp, init_vel]).flatten()

        y0, _, _ = self._calc_periodic_solution(drive_freq, drive_amp, y0_guess)

        def _solve_one_period(y0):               
            sol = diffrax.diffeqsolve(
                terms=diffrax.ODETerm(self._rhs),
                solver=diffrax.Kvaerno5(),
                t0=self.t0, t1=self.t1, dt0=None, max_steps=self.max_steps,
                y0=y0,
                saveat=diffrax.SaveAt(ts=jnp.linspace(self.t0, self.t1, self.n_time_steps)),
                throw=False,
                progress_meter=diffrax.NoProgressMeter(),
                stepsize_controller=diffrax.PIDController(rtol=self.rtol, atol=self.atol),
                args=(drive_amp, drive_freq),
            )
            return sol.ts, sol.ys

        ts, ys = jax.lax.cond(
            jnp.isnan(y0).any(),
            lambda: (jnp.zeros((self.n_time_steps,)),
                     jnp.zeros((self.n_time_steps, self.model.n_modes * 2))),
            lambda: _solve_one_period(y0)
        )

        return ts, ys
    
    def frequency_sweep(self,
             drive_freq: jax.Array,
             drive_amp: jax.Array, 
             sweep_direction: const.SweepDirection,
            ):

        drive_freq_mesh, drive_amp_mesh, init_disp_mesh, init_vel_mesh = self.multistart.generate_simulation_grid(
            self.model, drive_freq, drive_amp
        )
        drive_freq_mesh_flat = drive_freq_mesh.ravel()                       # (n_sim,)
        drive_amp_mesh_flat  = drive_amp_mesh.ravel()                        # (n_sim,)

        y0_guess = jnp.stack([init_disp_mesh, init_vel_mesh], axis=-1)  # (F, A, D, V, 2)
        y0_guess = y0_guess.reshape(-1, 2)                              # (n_sim, 2) for 1 mode

        periodic_solutions = jax.vmap(self._calc_periodic_solution)(
            drive_freq_mesh_flat, drive_amp_mesh_flat, y0_guess
        )  # y0: (n_sim, n_modes*2)  y_max: (n_sim, n_modes)

        return periodic_solutions

    @filter_jit
    def _calc_periodic_solution(self,
            driving_frequency: jax.Array,
            driving_amplitude: jax.Array,
            y0_guess: jax.Array):
        
        ts = jnp.linspace(self.t0, self.t1, self.m_segments + 1)
    
        s0 = diffrax.diffeqsolve(
            terms=diffrax.ODETerm(self._rhs), solver=diffrax.Kvaerno5(),
            t0=self.t0, t1=self.t1, dt0=None,
            y0=y0_guess,
            saveat=diffrax.SaveAt(ts=ts),
            throw=False,
            max_steps=self.max_steps,
            progress_meter=diffrax.NoProgressMeter(),
            stepsize_controller=diffrax.PIDController(rtol=self.rtol, atol=self.atol),
            args=(driving_amplitude, driving_frequency),
        ).ys[:-1]  # (m_segments, n_modes * 2)

        @filter_jit
        def _integrate_segment(sk, t0k, t1k):
            solk = diffrax.diffeqsolve(
                terms=diffrax.ODETerm(self._rhs),
                solver=diffrax.Kvaerno5(),
                t0=t0k, t1=t1k, dt0=None,
                y0=sk,
                adjoint=diffrax.RecursiveCheckpointAdjoint(),
                saveat=diffrax.SaveAt(t1=True),
                throw=False,
                max_steps=self.max_steps,
                progress_meter=diffrax.NoProgressMeter(),
                stepsize_controller=diffrax.PIDController(rtol=self.rtol, atol=self.atol),
                args=(driving_amplitude, driving_frequency),
            )
            return solk.ys.squeeze(0)   # (n_modes * 2,)

        # def _residual(s, _args=None):               # s: (m_segments, n_modes * 2)
        #     def _one(carry, i):
        #         sk  = s[i]
        #         sk1 = s[(i + 1) % self.m_segments]    # wrap for periodicity
        #         Phi = _integrate_segment(sk, ts[i], ts[i + 1])  # (n_modes * 2,)
        #         r = Phi - sk1
        #         return carry, r

        #     _, Rs = jax.lax.scan(_one, None, jnp.arange(self.m_segments))  # Rs: (m_segments, n_modes * 2)
        #     return Rs.reshape((-1,))                         # (m_segments * n_modes * 2,)
        
        def _residual(s, _args=None): # s: (m_segments + 1, n_modes * 2)
            t0 = ts[:-1]
            t1 = ts[1:]
            s_next = jnp.roll(s, shift=-1, axis=0)

            Phi = jax.vmap(_integrate_segment, in_axes=(0, 0, 0))(s, t0, t1)  # (m, n)
            R = Phi - s_next                                                  # (m, n)
            return R.reshape(-1)                                              # (m*n,)

        solver = optx.LevenbergMarquardt(rtol=1e-6, atol=1e-8, norm=optx.rms_norm) # for debugging: verbose=frozenset({"step", "accepted", "loss", "step_size"})
        sol = optx.least_squares(_residual, solver, y0=s0, options={"jac": "bwd"}, max_steps=self.max_shooting_iterations, throw=False) 

        # This again is probably very sensitive to the initial conditions again, have to think about that
        def _postprocess_success(s_star):
            y0_periodic = s_star[0]

            # fine trajectory to get x_max
            sol_traj = diffrax.diffeqsolve(
                terms=diffrax.ODETerm(self._rhs), solver=diffrax.Kvaerno5(),
                t0=self.t0, t1=self.t1, dt0=None,
                y0=y0_periodic,
                saveat=diffrax.SaveAt(ts=jnp.linspace(self.t0, self.t1, self.n_time_steps)),
                throw=False,
                max_steps=self.max_steps,
                progress_meter=diffrax.NoProgressMeter(),
                stepsize_controller=diffrax.PIDController(rtol=self.rtol, atol=self.atol),
                args=(driving_amplitude, driving_frequency),
            )
            xs  = sol_traj.ys[:, :self.model.n_modes]
            x_max = jnp.max(jnp.abs(xs))

            # Floquet multipliers from monodromy (via jacobian of flow)
            mu = self._compute_floquet_multipliers(y0_periodic, driving_amplitude, driving_frequency)

            # classify
            stable, bif_code, rho_max = self._classify_multipliers(mu)

            return {
                "y0": y0_periodic,                 # (2*n_modes,)
                "x_max": x_max,                    # scalar
                "mu": mu,                          # (2*n_modes,) complex
                "rho_max": rho_max,                # scalar
                "stable": stable,                  # bool
                "bifurcation": bif_code,           # int32
            }
            
        def _postprocess_fail(_):
            return {
                "y0": jnp.full((self.model.n_modes*2,), jnp.nan),
                "x_max": jnp.nan,
                "mu": jnp.full((self.model.n_modes*2,), jnp.nan + 0j),
                "rho_max": jnp.nan,
                "stable": False,
                "bifurcation": jnp.array(0, dtype=jnp.int32),
            }
        
        out = jax.lax.cond(
            sol.result == optx.RESULTS.successful,
            lambda: _postprocess_success(sol.value),
            lambda: _postprocess_fail(None)
        )

        return out

    @filter_jit
    def _compute_floquet_multipliers(self, y0, driving_amplitude, driving_frequency):
        """
        Compute Floquet multipliers using automatic differentiation.
        
        The monodromy matrix M is the Jacobian of the flow map Φ(T, y0) with respect to y0.
        The Floquet multipliers are the eigenvalues of M.
        """
        
        # Define the flow map for one period
        def flow_map(y):
            sol = diffrax.diffeqsolve(
                terms=diffrax.ODETerm(self._rhs),
                solver=diffrax.Kvaerno5(),
                t0=self.t0, t1=self.t1, dt0=None,
                adjoint=diffrax.ForwardMode(),
                y0=y,
                saveat=diffrax.SaveAt(t1=True),
                throw=False,
                max_steps=self.max_steps,
                progress_meter=diffrax.NoProgressMeter(),
                stepsize_controller=diffrax.PIDController(rtol=self.rtol, atol=self.atol),
                args=(driving_amplitude, driving_frequency),
            )
            return sol.ys.squeeze(0)
        
        # Compute the monodromy matrix using JAX's automatic differentiation
        monodromy_matrix = jax.jacfwd(flow_map)(y0)
        
        # Compute eigenvalues (Floquet multipliers)
        # Note: For real systems, eigenvalues might be complex
        eigenvalues, _ = jnp.linalg.eig(monodromy_matrix)
        
        return eigenvalues
    
    def _classify_multipliers(self, mu,
                            tol_inside=1e-6,   # margin inside unit circle for "stable"
                            tol_sn=1e-3,       # |mu - 1| small
                            tol_pd=1e-3,       # |mu + 1| small
                            tol_ns_r=1e-3,     # ||mu|-1| small
                            tol_ns_ang=0.2):   # angle not near 0 or pi
        # mu: (n,) complex
        rho = jnp.abs(mu)
        rho_max = jnp.max(rho)

        # stable if strictly inside by a little margin
        stable = rho_max <= (1.0 - tol_inside)

        # saddle-node: multiplier near +1
        sn = jnp.any(jnp.abs(mu - (1.0 + 0j)) < tol_sn)

        # period doubling: multiplier near -1
        pd = jnp.any(jnp.abs(mu + (1.0 + 0j)) < tol_pd)

        # Neimark–Sacker: complex pair near unit circle with nontrivial angle
        ang = jnp.angle(mu)
        complex_mask = (jnp.abs(jnp.imag(mu)) > 1e-9)
        near_circle  = jnp.abs(rho - 1.0) < tol_ns_r
        ang_nontrivial = (jnp.abs(jnp.mod(ang, jnp.pi)) > tol_ns_ang)
        ns = jnp.any(complex_mask & near_circle & ang_nontrivial)

        # 0:none/unknown  1:SN  2:PD  3:NS
        bif_code = jnp.where(sn, 1,
                    jnp.where(pd, 2,
                        jnp.where(ns, 3, 0))).astype(jnp.int32)

        return stable, bif_code, rho_max
    
    @filter_jit
    def _rhs(self, tau, y, args):
        _drive_amp, drive_freq  = args

        t = tau / drive_freq
        dydt = self.model.f(t, y, args) / drive_freq

        return dydt
    
    @filter_jit
    def _aug_rhs(self, tau, y_aug, args):
        _drive_amp, drive_freq = args
        
        y  = y_aug[:self.model.n_modes*2]
        X  = y_aug[self.model.n_modes*2:].reshape(self.model.n_modes, self.model.n_modes)

        t   = tau / drive_freq
        f   = self.model.f(t, y, args)
        f_y = self.model.f_y(t, y, args)

        dydt  = f / drive_freq
        dXdt  = ((f_y @ X) / drive_freq).reshape(-1)
        return jnp.hstack([dydt, dXdt])



from .abstract_sweep import AbstractSweep
from .sweep_directions import AbstractSweepDirection
from .sweep_directions import Forward, Backward
import jax.numpy as jnp
from jax import lax

class NearestNeighbourSweep(AbstractSweep):
    def __init__(self, sweep_direction: AbstractSweepDirection | list[AbstractSweepDirection] = [Forward(), Backward()]):
        super().__init__()
        self.sweep_direction = sweep_direction

    def to_dtype(self, dtype):
        return NearestNeighbourSweep(sweep_direction=self.sweep_direction)

    def sweep(self, periodic_solutions):
        """
        Expects periodic_solutions['max_x_total'] with shape:
        (n_freq, n_amp, n_init_disp, n_init_vel) and entries being scalar
        response amplitudes for each (init_disp, init_vel) seed.
        Returns:
          sweeped_vals: (n_freq, n_amp) chosen amplitude per freq & amp
          sweeped_idx:  (n_freq, n_amp) flat index of the chosen seed (debug)
        """
        max_x_total = periodic_solutions['max_x_total']  # scalar amplitudes
        n_freq, n_amp, _, _ = max_x_total.shape

        results = {'forward': None, 'backward': None}

        for direction_obj in (self.sweep_direction if isinstance(self.sweep_direction, list) else [self.sweep_direction]):
            if isinstance(direction_obj, Forward):
                order = jnp.arange(n_freq)
                start_idx = 0
                direction = 'forward'
            elif isinstance(direction_obj, Backward):
                order = jnp.arange(n_freq - 1, -1, -1)
                start_idx = n_freq - 1
                direction = 'backward'
            else:
                raise ValueError("Unsupported sweep direction")

            def sweep_one_amplitude(amp_idx):
                # candidates at the starting frequency, flattened over all seeds
                start_cands = max_x_total[start_idx, amp_idx].reshape(-1)

                # choose a reasonable start: the largest finite response (or first finite)
                finite_mask = jnp.isfinite(start_cands)
                # if all NaN/inf, default to zero
                start_choice = jnp.where(
                    finite_mask.any(),
                    start_cands[jnp.argmax(jnp.where(finite_mask, start_cands, -jnp.inf))],
                    0.0,
                )
                start_choice_idx = jnp.where(
                    finite_mask.any(),
                    jnp.argmax(jnp.where(finite_mask, start_cands, -jnp.inf)),
                    -1,
                )

                def body(carry, freq_idx):
                    prev_val, out_vals, out_idx = carry
                    cands = max_x_total[freq_idx, amp_idx].reshape(-1)  # (n_seeds,)
                    # distance in *amplitude* space
                    diffs = jnp.abs(cands - prev_val)

                    # ignore NaNs/Infs by giving them infinite distance
                    finite_mask = jnp.isfinite(cands)
                    diffs = jnp.where(finite_mask, diffs, jnp.inf)

                    k = jnp.argmin(diffs)
                    no_valid = jnp.logical_not(jnp.any(finite_mask))
                    chosen = jnp.where(no_valid, prev_val, cands[k])
                    chosen_idx = jnp.where(no_valid, -1, k)

                    out_vals = out_vals.at[freq_idx].set(chosen)
                    out_idx = out_idx.at[freq_idx].set(chosen_idx)
                    return (chosen, out_vals, out_idx), None

                init_vals = jnp.full((n_freq,), jnp.nan)
                init_idx = jnp.full((n_freq,), -1, dtype=jnp.int32)

                (final_val, out_vals, out_idx), _ = lax.scan(
                    body, (start_choice, init_vals, init_idx), order
                )
                return out_vals, out_idx

            vals, idxs = zip(*(sweep_one_amplitude(amp) for amp in range(n_amp)))
            sweeped_vals = jnp.stack(vals, axis=1)  # (n_freq, n_amp)
            results[direction] = sweeped_vals

        return results



import matplotlib.pyplot as plt
import jax.numpy as jnp
import numpy as np
from matplotlib import cm, colors

from .. import constants as const

BIF_LABELS = {0: "none", 1: "SN", 2: "PD", 3: "NS"}

def get_data():

    # Load nonlinear sweep data
    file = f"tests/chris/16_40_15-LDV_100-nonlinear_sweep2_data.txt"
    data = np.loadtxt(file, delimiter=",", skiprows=1)  # Skip the header row
    freq = data[:, 0]
    mag = data[:, 1]
    phase = data[:, 2]

    def measurement2meter(freq, mag, setup, *args):
        if setup == "ENIGMA with correction":
            # ENIGMA - using transduction
            transduction = args[0]
            delta = mag / transduction * 1e-9
            gamma = 4*np.pi / 632.8e-9  # 4 pi / lambda_red
            lin_error = 1/8*delta*2 * gamma*2
            return delta * (1 - lin_error)

        elif setup == "ENIGMA":
            # ENIGMA - using transduction
            transduction = args[0]
            return mag / transduction * 1e-9

        elif setup == "Polytec":
            # Polytec - using LDV settings
            LDV = args[0]
            return 2 * mag * LDV / (2 * np.pi * freq)

        else:
            print(f'Setup should be "ENIGMA" or "ENIGMA with correction" or "Polytec", not "{setup}"')

    settings = "LDV_100"
    setup = "Polytec"; calibration = float(settings.split('_')[-1]) * 1e-3
    total_points = 11; picked_point = 2

    x_ref = 1e-8
    omega_ref = 207.708e3

    freq_sweep = freq
    modeshape_coeff1 = np.sin((picked_point - 1) / (total_points - 1) * np.pi)
    mag_sweep = measurement2meter(freq_sweep, mag, setup, calibration) / modeshape_coeff1

    freq_sweep = freq_sweep / omega_ref
    mag_sweep = mag_sweep / x_ref

    return freq_sweep, mag_sweep

def _truncate_cmap(name="Greys", lo=0.3, hi=1.0, N=256):
    base = cm.get_cmap(name, N)
    return colors.LinearSegmentedColormap.from_list(
        f"{name}_trunc", base(np.linspace(lo, hi, N))
    )

def _envelope_by_bins(f, y, *, mask=None, nbins=300, agg="max"):
    """
    Build a single-valued 'envelope' y_env(f_env) from scattered (f, y) by binning.
    agg: 'max' (upper branch) or 'median' (central tendency).
    """
    f = np.asarray(f).ravel()
    y = np.asarray(y).ravel()
    if mask is None:
        mask = np.isfinite(f) & np.isfinite(y)
    else:
        mask = mask & np.isfinite(f) & np.isfinite(y)

    if mask.sum() < 3:
        return np.array([]), np.array([])

    fmin, fmax = float(np.min(f[mask])), float(np.max(f[mask]))
    if not np.isfinite(fmin) or not np.isfinite(fmax) or fmin == fmax:
        return np.array([]), np.array([])

    edges = np.linspace(fmin, fmax, nbins + 1)
    idx = np.digitize(f[mask], edges) - 1  # bin indices in [0, nbins-1]
    f_env = []
    y_env = []
    for b in range(nbins):
        sel = idx == b
        if not np.any(sel):
            continue
        fbin = f[mask][sel]
        ybin = y[mask][sel]
        if agg == "median":
            yv = float(np.median(ybin))
        else:
            yv = float(np.max(ybin))
        f_env.append(0.5 * (edges[b] + edges[b + 1]))
        y_env.append(yv)
    return np.asarray(f_env), np.asarray(y_env)

def _weighted_stats(x, y, power=2.0):
    """
    Resonance-weighted mean and std of x using weights ~ y^power (y assumed >= 0).
    """
    x = np.asarray(x).ravel()
    y = np.asarray(y).ravel()
    m = np.isfinite(x) & np.isfinite(y)
    if m.sum() == 0:
        return np.nan, np.nan
    w = np.maximum(y[m], 0.0) ** power
    if w.sum() == 0:
        return np.nan, np.nan
    mu = np.sum(w * x[m]) / np.sum(w)
    var = np.sum(w * (x[m] - mu) ** 2) / np.sum(w)
    return mu, np.sqrt(var)

def _align_measurement_to_sim(f_meas, x_meas, f_env, y_env, *, power=2.0):
    """
    Find linear frequency map f' = s * f_meas + t and amplitude scale c
    so that x_meas aligns to y_env on the overlapping domain.
    """
    # Clean + sort
    f_meas = np.asarray(f_meas).ravel()
    x_meas = np.asarray(x_meas).ravel()
    m = np.isfinite(f_meas) & np.isfinite(x_meas)
    f_meas, x_meas = f_meas[m], x_meas[m]
    order = np.argsort(f_meas)
    f_meas, x_meas = f_meas[order], x_meas[order]

    f_env = np.asarray(f_env).ravel()
    y_env = np.asarray(y_env).ravel()
    m2 = np.isfinite(f_env) & np.isfinite(y_env)
    f_env, y_env = f_env[m2], y_env[m2]
    order2 = np.argsort(f_env)
    f_env, y_env = f_env[order2], y_env[order2]

    if f_meas.size < 3 or f_env.size < 3:
        # Fallback: identity mapping, unit amplitude
        return 1.0, 0.0, 1.0

    # Resonance-weighted centroid+width alignment (robust for unimodal resonance)
    mu_m, sd_m = _weighted_stats(f_meas, x_meas, power=power)
    mu_s, sd_s = _weighted_stats(f_env, y_env, power=power)
    if not np.isfinite(mu_m) or not np.isfinite(mu_s) or sd_m <= 0 or not np.isfinite(sd_s):
        s, t = 1.0, 0.0
    else:
        s = sd_s / sd_m if sd_m > 0 else 1.0
        t = mu_s - s * mu_m

    f_meas_mapped = s * f_meas + t

    # Compute amplitude scale c on the overlapping region by least squares
    # Interpolate measurement -> env grid
    within = (f_env >= f_meas_mapped.min()) & (f_env <= f_meas_mapped.max())
    if np.any(within):
        x_on_env = np.interp(f_env[within], f_meas_mapped, x_meas)
        y_tgt = y_env[within]
        finite = np.isfinite(x_on_env) & np.isfinite(y_tgt)
        if np.any(finite):
            denom = float(np.sum(x_on_env[finite] ** 2))
            c = float(np.sum(x_on_env[finite] * y_tgt[finite]) / denom) if denom > 0 else 1.0
        else:
            c = 1.0
    else:
        c = 1.0

    return s, t, c

def plot_branch_exploration(
    coarse_drive_freq_mesh,
    coarse_drive_amp_mesh,
    results,
    *,
    stable_size=60,
    unstable_size=16,
    tol_inside=1e-4,
    annotate_bifurcations=False,
    cmap_lo=0.15,
    cmap_hi=0.90,
    backbone=None,
    backbone_n=300,
    backbone_kwargs=None,
    title="Branch Exploration",
    overlay_get_data=True,
    align_measurement=True,     # <— auto-align frequency & amplitude so curves line up
    show_interpolated=False,     # <— also show markers at sim grid (one-to-one comparison)
    env_bins=300,               # <— envelope resolution for alignment/comparison
    env_agg="max",              # 'max' = upper branch, 'median' = central tendency
):
    # --- flatten grids
    freq_vals = jnp.asarray(coarse_drive_freq_mesh).ravel()
    amp_vals  = jnp.asarray(coarse_drive_amp_mesh).ravel()
    n_sim     = int(freq_vals.size)

    # helpers
    def _flat(name, default=None):
        if name in results and results[name] is not None:
            return jnp.asarray(results[name]).reshape(-1)
        return default

    x_max = _flat("x_max", default=jnp.full((n_sim,), jnp.nan))

    if "stable" in results:
        stable = _flat("stable")
    elif "rho_max" in results:
        rho_max = _flat("rho_max")
        stable = jnp.isfinite(rho_max) & (rho_max <= (1.0 - tol_inside))
    elif "mu" in results:
        mu = jnp.asarray(results["mu"]).reshape(n_sim, -1)
        rho = jnp.nanmax(jnp.abs(mu), axis=1)
        stable = jnp.isfinite(rho) & (rho <= (1.0 - tol_inside))
    else:
        stable = jnp.zeros((n_sim,), dtype=bool)

    sizes = jnp.where(stable, stable_size, unstable_size)
    bif = _flat("bifurcation", default=jnp.zeros((n_sim,), dtype=jnp.int32))

    # --- colormap
    cmap = _truncate_cmap("Greys", cmap_lo, cmap_hi)

    # --- color scale bounds
    vmin = float(amp_vals.min())
    vmax = float(amp_vals.max())
    if not np.isfinite(vmin) or not np.isfinite(vmax):
        vmin, vmax = 0.0, 1.0
    if vmin == vmax:
        eps = 1e-12
        vmin -= eps
        vmax += eps

    # --- figure
    fig, ax = plt.subplots(figsize=(8, 6))
    sc = ax.scatter(
        np.asarray(freq_vals),
        np.asarray(x_max),
        c=np.asarray(amp_vals),
        s=np.asarray(sizes),
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        edgecolors="none",
        alpha=0.9,
    )
    fig.colorbar(sc, ax=ax, label="Driving amplitude")

    # stability legend
    stable_dot   = ax.scatter([], [], s=stable_size,   c="k", alpha=0.9, label="stable")
    unstable_dot = ax.scatter([], [], s=unstable_size, c="k", alpha=0.9, label="unstable")
    legend_handles = [stable_dot, unstable_dot]

    # --- backbone (optional)
    if backbone is not None:
        f0   = float(backbone["f0"])
        beta = float(backbone["beta"])
        A_min = float(np.nanmin(np.asarray(x_max)))
        A_max = float(np.nanmax(np.asarray(x_max)))
        if not np.isfinite(A_min) or not np.isfinite(A_max):
            A_min, A_max = 0.0, 1.0
        A_min = max(0.0, A_min)

        A_grid = np.linspace(A_min, A_max, backbone_n)
        # k = 3/(16*np.pi**2)  # if your x-axis is in Hz
        k = 3/4               # if your x-axis is in rad/s
        inside = f0**2 + k * beta * A_grid**2
        mask = inside > 0.0
        f_grid = np.sqrt(inside[mask])
        A_plot = A_grid[mask]

        default_style = {"lw": 2, "label": "Backbone"}
        default_style.update(backbone_kwargs or {})
        (h_bb,) = ax.plot(f_grid, A_plot, **default_style)
        legend_handles.append(h_bb)

    # --- overlay measurement from get_data()
    if overlay_get_data:
        try:
            f_meas, x_meas = get_data()
            # Clean measurement
            f_meas = np.asarray(f_meas).ravel()
            x_meas = np.asarray(x_meas).ravel()
            m = np.isfinite(f_meas) & np.isfinite(x_meas)
            f_meas, x_meas = f_meas[m], x_meas[m]
            order = np.argsort(f_meas)
            f_meas, x_meas = f_meas[order], x_meas[order]

            # Build a simulation 'envelope' to compare against (use stable points if available)
            f_sim = np.asarray(freq_vals).ravel()
            x_sim = np.asarray(x_max).ravel()
            stab = np.asarray(stable).ravel().astype(bool) if stable is not None else None
            f_env, y_env = _envelope_by_bins(f_sim, x_sim, mask=stab, nbins=env_bins, agg=env_agg)

            # Decide alignment
            if align_measurement and f_env.size > 5:
                s, t, c = _align_measurement_to_sim(f_meas, x_meas, f_env, y_env, power=2.0)
            else:
                s, t, c = 1.0, 0.0, 1.0

            f_meas_mapped = s * f_meas + t
            x_meas_scaled = c * x_meas

            # 1) Raw aligned curve
            (h_meas_raw,) = ax.plot(
                f_meas_mapped,
                x_meas_scaled,
                lw=1.8,
                alpha=0.95,
                label="Measurement (aligned)",
            )
            legend_handles.append(h_meas_raw)

            # 2) Optional: markers at sim envelope frequencies (direct comparison)
            if show_interpolated and f_env.size > 0:
                # Only where we overlap
                overlap = (f_env >= f_meas_mapped.min()) & (f_env <= f_meas_mapped.max())
                if np.any(overlap):
                    x_on_env = np.interp(f_env[overlap], f_meas_mapped, x_meas_scaled)
                    (h_interp,) = ax.plot(
                        f_env[overlap],
                        x_on_env,
                        ls="none",
                        marker="x",
                        ms=4,
                        alpha=0.9,
                        label="Measurement (on sim grid)",
                    )
                    legend_handles.append(h_interp)

                # Optional: show the sim envelope itself for reference (thin line)
                (h_env,) = ax.plot(
                    f_env, y_env, lw=1.0, ls="--", alpha=0.6, label="Sim envelope"
                )
                legend_handles.append(h_env)

        except Exception as e:
            print(f"Could not overlay get_data(): {e}")

    # --- optional bifurcation overlays
    if annotate_bifurcations:
        sn_mask = np.asarray(bif == 1)
        pd_mask = np.asarray(bif == 2)
        ns_mask = np.asarray(bif == 3)

        if sn_mask.any():
            h_sn = ax.scatter(
                np.asarray(freq_vals)[sn_mask],
                np.asarray(x_max)[sn_mask],
                marker="o", facecolors="none", edgecolors="C3", s=80, linewidths=1.5, label="SN",
            ); legend_handles.append(h_sn)

        if pd_mask.any():
            h_pd = ax.scatter(
                np.asarray(freq_vals)[pd_mask],
                np.asarray(x_max)[pd_mask],
                marker="s", facecolors="none", edgecolors="C0", s=80, linewidths=1.5, label="PD",
            ); legend_handles.append(h_pd)

        if ns_mask.any():
            h_ns = ax.scatter(
                np.asarray(freq_vals)[ns_mask],
                np.asarray(x_max)[ns_mask],
                marker="^", facecolors="none", edgecolors="C2", s=80, linewidths=1.5, label="NS",
            ); legend_handles.append(h_ns)

    if legend_handles:
        ax.legend(handles=legend_handles, title="Legend")

    ax.set_xlabel("Driving frequency")
    ax.set_ylabel("Steady-state displacement amplitude")
    ax.set_title(title)
    ax.grid(const.PLOT_GRID)
    plt.tight_layout()
    plt.savefig("branch_exploration.png", dpi=300)
    plt.show()


import math
import jax
import jax.numpy as jnp
from equinox import filter_jit
import diffrax
from jax import core as jax_core

from .abstract_solver import AbstractSolver
from ..oscillator.abstract_oscillator import AbstractOscillator
from ..multistart.abstract_multistart import AbstractMultistart
from ..excitation.abstract_excitation import AbstractExcitation
from ..sweep.abstract_sweep import AbstractSweep
from ..result.frequency_sweep_result import FrequencySweepResult

from .. import constants as const 

class TimeIntegrationSolver(AbstractSolver):
    def __init__(self, rtol: float = 1e-4, atol: float = 1e-7, n_time_steps: int = None, max_steps: int = 4096, 
                 verbose: bool = False, throw: bool = False):
        
        self.max_steps = max_steps
        self.n_time_steps = n_time_steps
        self.rtol = rtol
        self.atol = atol
        self.verbose = verbose
        self.throw = throw

        self.model: AbstractOscillator = None
        self.multistarter: AbstractMultistart = None

    @staticmethod
    def _is_tracer(value) -> bool:
        """Check whether a value is being traced by JAX (e.g. inside vmap/jit)."""
        return isinstance(value, jax_core.Tracer)
    
    def _max_steps_budget(self, t_span: float, period: float, safety_factor: float = 2.0) -> int:
        """Infer a max_steps large enough for the requested horizon.

        We approximate the solver step size with the sampling interval
        (period / (n_time_steps * MAXIMUM_ORDER_SUPERHARMONICS)) and
        inflate by a safety factor so we don't trip the diffrax max_steps
        guard on long, low-frequency runs.
        """
        if self.n_time_steps is None:
            return self.max_steps

        # When traced (e.g. under vmap/jit), fall back to the configured cap
        # instead of trying to convert tracers to Python scalars.
        if self._is_tracer(t_span) or self._is_tracer(period):
            return self.max_steps

        steps_per_period = max(int(self.n_time_steps), 1) * const.MAXIMUM_ORDER_SUPERHARMONICS
        dt_est = float(period) / steps_per_period
        if not math.isfinite(dt_est) or dt_est <= 0.0:
            return self.max_steps

        est_steps = math.ceil((t_span / dt_est) * safety_factor)
        return max(self.max_steps, est_steps)

    def time_response(self,
                 f_omega: jax.Array,  
                 f_amp: jax.Array, 
                 x0: jax.Array,  
                 v0: jax.Array,
                **kwargs
                ):

        y0 = jnp.concatenate([jnp.atleast_1d(x0), jnp.atleast_1d(v0)], axis=-1)

        if self.n_time_steps is None and not self._is_tracer(f_omega):
            rtol = 0.01
            max_frequency_component = const.MAXIMUM_ORDER_SUPERHARMONICS * jnp.max(f_omega)
            
            one_period = 2.0 * jnp.pi / max_frequency_component
            sampling_frequency = jnp.pi / (jnp.sqrt(2 * rtol)) * max_frequency_component
            
            n_time_steps = int(math.ceil(float(one_period * sampling_frequency)))
            self.n_time_steps = n_time_steps
        elif self.n_time_steps is None:
            raise ValueError("n_time_steps must be set before calling time_response when tracing.")

        period = jnp.max(2.0 * jnp.pi / f_omega)
        periods_to_retain = const.N_PERIODS_TO_RETAIN
        T = period * periods_to_retain
        t_ss = jnp.max(self.model.t_steady_state(f_omega * 2.0 * jnp.pi, ss_tol=self.rtol)) * const.SAFETY_FACTOR_T_STEADY_STATE
        t0 = 0.0
        t1 = t_ss + T

        if kwargs.get("only_save_steady_state"):
            ts = jnp.linspace(t_ss, t1, self.n_time_steps * 10)
        else:
            n_periods = (t1 - t0) / T
            if self._is_tracer(n_periods):
                n_time_steps_total = self.n_time_steps * periods_to_retain
            else:
                n_time_steps_total = self.n_time_steps * int(math.ceil(float(n_periods)))
            ts = jnp.linspace(t0, t1, n_time_steps_total)

        max_steps_budget = self._max_steps_budget(t1 - t0, period)

        sol = diffrax.diffeqsolve(
                terms=diffrax.ODETerm(self._rhs),
                solver=diffrax.Tsit5(),
                t0=t0, t1=t1, dt0=None, max_steps=max_steps_budget,
                y0=y0,
                saveat=diffrax.SaveAt(ts=ts),
                throw=self.throw,
                progress_meter=diffrax.NoProgressMeter(),
                stepsize_controller=diffrax.PIDController(rtol=self.rtol, atol=self.atol),
                args=(f_amp, f_omega),
        )

        return sol.ts, sol.ys
        
    # TO DO: Add phase difference results
    def frequency_sweep(self,
             excitor: AbstractExcitation,
             sweeper: AbstractSweep,
            ) -> FrequencySweepResult:
        
        periods_to_retain = const.N_PERIODS_TO_RETAIN

        @filter_jit
        def solve_one_case(f_omega, f_amp, x0, v0):  
            x0 = jnp.full((self.model.n_modes,), x0)         
            v0 = jnp.full((self.model.n_modes,), v0)
            y0 = jnp.concatenate([jnp.atleast_1d(x0), jnp.atleast_1d(v0)], axis=-1)

            period = jnp.max(2.0 * jnp.pi / f_omega)
            T = period * periods_to_retain
            t_ss = jnp.max(self.model.t_steady_state(f_omega, ss_tol=self.rtol)) * const.SAFETY_FACTOR_T_STEADY_STATE
            
            t0 = 0.0
            t1 = t_ss + T
            ts = jnp.linspace(t_ss, t1, self.n_time_steps * periods_to_retain)
            
            sol = diffrax.diffeqsolve(
                terms=diffrax.ODETerm(self._rhs),
                solver=diffrax.Tsit5(),
                t0=t0, t1=t1, dt0=None, max_steps=max_steps_budget,
                y0=y0,
                saveat=diffrax.SaveAt(ts=ts),
                throw=self.throw,
                progress_meter=diffrax.NoProgressMeter(),
                stepsize_controller=diffrax.PIDController(rtol=self.rtol, atol=self.atol),
                args=(f_amp, f_omega),
            )

            # Treat any non-finite trajectories as failures to avoid polluting sweeps
            is_finite = jnp.all(jnp.isfinite(sol.ys))
            successful = jnp.logical_and(
                sol.result == diffrax.RESULTS.successful,
                is_finite,
            )

            xs = sol.ys[:, :self.model.n_modes]
            vs = sol.ys[:, self.model.n_modes:]

            max_x_total = jnp.where(
                successful,
                jnp.max(jnp.abs(jnp.sum(xs, axis=-1))),
                jnp.nan,
            )
            max_v_total = jnp.where(
                successful,
                jnp.max(jnp.abs(jnp.sum(vs, axis=-1))),
                jnp.nan,
            )

            max_x_modes = jnp.where(
                successful,
                jnp.max(jnp.abs(xs), axis=0),
                jnp.full_like(xs[0], jnp.nan),
            )
            max_v_modes = jnp.where(
                successful,
                jnp.max(jnp.abs(vs), axis=0),
                jnp.full_like(vs[0], jnp.nan),
            )

            return dict(
                f_omega=f_omega, 
                f_amp=f_amp,
                x0=x0, 
                v0=v0, 
                max_x_total=max_x_total, 
                max_v_total=max_v_total, 
                max_x_modes=max_x_modes, 
                max_v_modes=max_v_modes, 
                successful=successful
            )
            
        f_omegas = excitor.f_omegas
        f_amps = excitor.f_amps
        
        # TO DO: Check if this is appropriate
        if self.n_time_steps is None:
            if self._is_tracer(f_omegas):
                raise ValueError("n_time_steps must be set before calling frequency_sweep when tracing.")
            rtol = 0.01
            max_frequency_component = const.MAXIMUM_ORDER_SUPERHARMONICS * jnp.max(f_omegas)
            
            one_period = 2.0 * jnp.pi / max_frequency_component
            sampling_frequency = jnp.pi / (jnp.sqrt(2 * rtol)) * max_frequency_component

            n_time_steps = int(math.ceil(float(one_period * sampling_frequency)))
            self.n_time_steps = n_time_steps
        
        f_omegas, f_amps, x0s, v0s, shape = self.multistarter.generate_simulation_grid(self.model, f_omegas, f_amps)
        longest_period = jnp.max(2.0 * jnp.pi / f_omegas)
        t_ss_estimate = jnp.max(self.model.t_steady_state(f_omegas, ss_tol=self.rtol) * const.SAFETY_FACTOR_T_STEADY_STATE)
        t_span_estimate = t_ss_estimate + longest_period * periods_to_retain
        max_steps_budget = self._max_steps_budget(t_span_estimate, longest_period)

        flat_solutions = jax.vmap(solve_one_case, in_axes=(0, 0, 0, 0))(f_omegas, f_amps, x0s, v0s)

        periodic_solutions = jax.tree_util.tree_map(
            lambda leaf: leaf.reshape(shape[:-1] + leaf.shape[1:]),
            flat_solutions
        )

        successful_mask = periodic_solutions["successful"]
        n_successful = jnp.count_nonzero(successful_mask)
        n_total = successful_mask.size
        success_rate = jnp.where(n_total > 0, n_successful / n_total, 0.0)
        if not self._is_tracer(n_successful):
            n_successful = int(n_successful)
        if not self._is_tracer(success_rate):
            success_rate = float(success_rate)
        
        sweeped_periodic_solutions = sweeper.sweep(periodic_solutions)
                
        result = FrequencySweepResult(
            f_omegas=excitor.f_omegas,
            f_amps=excitor.f_amps,
            modal_forces=excitor.modal_forces,
            Q=self.model.Q,
            omega_0=self.model.omega_0,
            alpha=self.model.alpha,
            gamma=self.model.gamma,
            periodic_solutions=periodic_solutions,
            sweeped_periodic_solutions=sweeped_periodic_solutions,
            n_successful=n_successful,
            n_total=n_total,
            success_rate=success_rate,
        )

        return result

    @filter_jit
    def _rhs(self, t, y, args):
        f_amp, f_omega = args

        dy_dt = self.model.f(t, y, (f_amp, f_omega))

        return dy_dt



import jax.numpy as jnp
from jax import tree_util

class OneToneExcitation:
    def __init__(self, drive_frequencies, drive_amplitudes, modal_forces):
        self.drive_frequencies = drive_frequencies
        self.drive_amplitudes = drive_amplitudes
        self.modal_forces = modal_forces
                
        self.f_omegas = jnp.asarray(drive_frequencies)
        self.f_amps = jnp.outer(jnp.asarray(drive_amplitudes), jnp.asarray(modal_forces))
        
    def to_dtype(self, dtype):
        return OneToneExcitation(
            drive_frequencies=self.drive_frequencies.astype(dtype),
            drive_amplitudes=self.drive_amplitudes.astype(dtype),
            modal_forces=self.modal_forces.astype(dtype)
        )

# Register OneToneExcitation as a pytree
def _tree_flatten(obj):
    leaves = (obj.f_omegas, obj.f_amps)
    aux_data = (obj.drive_frequencies, obj.drive_amplitudes, obj.modal_forces)
    return leaves, aux_data

def _tree_unflatten(aux_data, leaves):
    f_omegas, f_amps = leaves
    drive_frequencies, drive_amplitudes, modal_forces = aux_data
    obj = OneToneExcitation(drive_frequencies, drive_amplitudes, modal_forces)
    obj.f_omegas = f_omegas
    obj.f_amps = f_amps
    return obj

tree_util.register_pytree_node(OneToneExcitation, _tree_flatten, _tree_unflatten)


# constants.py
from enum import Enum, auto
import jax

class SweepDirection(Enum):
    FORWARD  = auto()
    BACKWARD = auto()

class Precision(Enum):
    SINGLE = auto()
    DOUBLE = auto()

class ResponseType(Enum):
    FrequencyResponse = auto()
    TimeResponse      = auto()

MAXIMUM_ORDER_SUBHARMONICS = 10
MAXIMUM_ORDER_SUPERHARMONICS = 10

DT_MIN_FACTOR = 2000
DT_MAX_FACTOR = 200

MAX_SHOOTING_ITERATIONS = 20

PLOT_GRID = True

N_PERIODS_TO_RETAIN = 10
SAFETY_FACTOR_T_STEADY_STATE = 1.2



import jax.numpy as jnp
from equinox import filter_jit

class LagrangeBasis:
    '''
    This class forms a lagrange basis of degree m. Barycentric weights are not used, because
    the lagrange polynomials are only evaluated at the collocation points. We initialize with the collocation
    points in [0,1] so we can then easily evaluate p and dp_dt at the collocation points.    
    '''

    def __init__(self, m: int, t: jnp.ndarray):
        self.m = m
        self.t = jnp.atleast_1d(t)
        self.nodes = jnp.linspace(0.0, 1.0, self.m + 1)
        
        self.Phi = self._lagrange_basis(self.t)
        self.dPhi_dt = self._derivative_lagrange_basis(self.t)

    @filter_jit
    def evaluate(self, Y: jnp.ndarray) -> jnp.ndarray:
        '''
        Evaluate the Lagrange basis polynomials and their derivatives at the given points.

        Parameters:
        Y: jnp.ndarray
            Coefficients for the Lagrange basis polynomials, shape (m+1, n)
        Returns:
        p: jnp.ndarray
            Evaluated polynomials at the collocation points, shape (m, n)
        dp_dt: jnp.ndarray
            Evaluated derivatives at the collocation points, shape (m, n)
        '''
        p = self.Phi @ Y
        dp_dt = self.dPhi_dt @ Y
        return p, dp_dt 

    @filter_jit
    def _lagrange_basis(self, t: jnp.ndarray) -> jnp.ndarray: 
        M = self.nodes.shape[0]

        diff_all = t[:, None] - self.nodes[None, :]
        P = jnp.prod(diff_all, axis=1, keepdims=True)

        dn = self.nodes[:, None] - self.nodes[None, :]
        mask = ~jnp.eye(M, dtype=bool)
        D = jnp.prod(jnp.where(mask, dn, 1.0), axis=1)

        denom = diff_all * D[None, :]
        Phi = P / denom 
        return Phi

    @filter_jit
    def _derivative_lagrange_basis(self, t: jnp.ndarray) -> jnp.ndarray:
        Phi = self._lagrange_basis(t)
        diff_all = t[:, None] - self.nodes[None, :]

        inv = 1.0 / diff_all
        sum_all = jnp.sum(inv, axis=1, keepdims=True)
        Ssum = sum_all - inv

        dPhi = Phi * Ssum
        return dPhi
    
if __name__ == "__main__":
    import numpy as np

    m = 3

    # Use midpoints between nodes to avoid division-by-zero at nodes
    nodes = jnp.linspace(0.0, 1.0, m + 1)
    t_eval = 0.5 * (nodes[:-1] + nodes[1:])  # shape (m,)

    lagrange_basis = LagrangeBasis(m, t_eval)

    Y = jnp.array([[1.0], [2.0], [3.0], [4.0]])  # shape (m+1, 1)

    p, dp_dt = lagrange_basis.evaluate(Y)

    print("t_eval:", t_eval)
    print("Lagrange Basis Evaluation:")
    print("p:", p)
    print("dp/dt:", dp_dt)




# Copyright (c) 2026 Raymond Knetemann
# Licensed under the MIT License. See LICENSE file for details.

from __future__ import annotations
import jax
import jax.numpy as jnp
from typing import Tuple, Dict

from .oscillator.abstract_oscillator import AbstractOscillator
from .solver.abstract_solver import AbstractSolver
from .excitation.abstract_excitation import AbstractExcitation
from .multistart.abstract_multistart import AbstractMultistart
from .sweep.abstract_sweep import AbstractSweep
from .result.frequency_sweep_result import FrequencySweepResult 

from .solver.time_integration_solver import TimeIntegrationSolver
from .multistart.linear_response_multistart import LinearResponseMultistart
from .sweep.nearest_neighbour_sweep import NearestNeighbourSweep

from . import constants as const

def frequency_sweep(
    model: AbstractOscillator,
    excitor: AbstractExcitation,
    sweeper: AbstractSweep = NearestNeighbourSweep(),
    solver: AbstractSolver = TimeIntegrationSolver(),
    multistarter: AbstractMultistart = LinearResponseMultistart(),
    precision: const.Precision = const.Precision.SINGLE,
) -> FrequencySweepResult:
    """Run a frequency sweep for a dynamical model.

    Args:
        model: ...
        excitor: ...
        sweeper: ...
        solver: ...
        multistarter: ...
        precision: ...

    Returns:
        A `FrequencySweepResult` containing the sweep response.
    """
            
    if precision == const.Precision.DOUBLE:
        jax.config.update("jax_enable_x64", True)
        dtype = jnp.float64
    elif precision == const.Precision.SINGLE:
        jax.config.update("jax_enable_x64", False)
        dtype = jnp.float32
    else:
        raise ValueError(f"Unsupported precision: {precision}")

    model = model.to_dtype(dtype)
    excitor = excitor.to_dtype(dtype)
    sweeper = sweeper.to_dtype(dtype)
    multistarter = multistarter.to_dtype(dtype)
    
    solver.model = model
    solver.multistarter = multistarter

    frequency_sweep = solver.frequency_sweep(excitor, sweeper)

    return frequency_sweep


from .frequency_sweep import frequency_sweep
from .time_response import time_response

from .oscillator.nonlinear_oscillator import *
from .oscillator.van_der_pol_oscillator import *

from .solver.abstract_solver import *
from .solver.shooting_solver import *
from .solver.collocation_solver import *
from .solver.time_integration_solver import *

from .sweep.sweep_directions import *
from .sweep.nearest_neighbour_sweep import *

from .multistart.linear_response_multistart import *

from .excitation.one_tone import *

from .utils.plotting import *

from .constants import *

"""
A Python toolkit for simulating and visualizing nonlinear oscillators using experimentally realistic setups, supporting both time- and frequency-domain analyses.
"""


from __future__ import annotations
import jax
import jax.numpy as jnp
from functools import partial
from typing import Optional

from jax import tree_util

from .. import constants as const
from .abstract_oscillator import AbstractOscillator, oscillator

@oscillator
class NonlinearOscillator(AbstractOscillator):
    Q: jax.Array
    omega_0: jax.Array
    alpha: jax.Array
    gamma: jax.Array

    def __post_init__(self):
        jnp.asarray(self.Q)
        jnp.asarray(self.omega_0)
        jnp.asarray(self.alpha)
        jnp.asarray(self.gamma)

    def f(self, tau, state, args, omega_ref=1.0, x_ref=1.0):
        q, dq_dtau   = jnp.split(state, 2)
        f_amp, f_omega = [jnp.asarray(v).squeeze(()) for v in args]

        #omega_ref = f_omega[0]

        damping_term = (self.omega_0/omega_ref) * 1/self.Q * dq_dtau
        linear_stiffness_term = (1/omega_ref**2) * self.omega_0**2 * q
        quadratic_stiffness_term = (x_ref / omega_ref**2) * jnp.einsum("ijk,j,k->i", self.alpha, q, q)
        cubic_stiffness_term = (x_ref**2 / omega_ref**2) * jnp.einsum("ijkl,j,k,l->i", self.gamma, q, q, q) # Shape: (n_modes,)
        forcing_term = f_amp / (omega_ref**2 * x_ref) * jnp.cos(f_omega/omega_ref * tau)

        d2q_dtau2 = (
            - damping_term
            - linear_stiffness_term
            - quadratic_stiffness_term
            - cubic_stiffness_term
            + forcing_term
        ) 
        return jnp.concatenate([dq_dtau, d2q_dtau2])

    def f_y(self, tau, state, args):
        q, dq_dtau = jnp.split(state, 2)

        zero_block = jnp.zeros((self.n_modes, self.n_modes))
        identity_block = jnp.eye(self.n_modes)
        A_bottom_left = -jnp.diag(self.g2 + 3 * self.g3 * q**2)
        A_bottom_right = -jnp.diag(self.g1)

        A = jnp.block([[zero_block, identity_block],
                       [A_bottom_left, A_bottom_right]])
        return A
    
    @property
    def n_modes(self) -> int:
        return self.Q.shape[0]
    
    @property
    def n_states(self) -> int:
        return self.n_modes * 2

    def t_steady_state(self, driving_frequency: jax.Array, ss_tol: float) -> float:
        '''driving_frequency
        Calculates the settling time for a given Q-factor and driving frequency.
        Equation from Eq.5.10b Vibrations 2nd edition by Balakumar Balachandran | Edward B. Magrab
        '''
        #driving_frequency = 1.0
        t_steady_state = jnp.max(-2 * jnp.max(self.Q) * jnp.log(ss_tol * jnp.sqrt(1 - 1 / (4 * jnp.max(self.Q)**2)) / (driving_frequency))).reshape(())

        return t_steady_state
    
    def to_dtype(self, dtype: jnp.dtype) -> NonlinearOscillator:
        return NonlinearOscillator(
            Q=self.Q.astype(dtype),
            omega_0=self.omega_0.astype(dtype),
            alpha=self.alpha.astype(dtype),
            gamma=self.gamma.astype(dtype)
        )
    
    def __repr__(self):
        Q_terms = ", ".join([f"Q[{i}]={float(v):.6f}" for i, v in enumerate(self.Q)])
        omega_0_terms = ", ".join([f"omega_0[{i}]={float(v):.6f}" for i, v in enumerate(self.omega_0)])

        alpha_indices, alpha_value = jnp.where(self.alpha != 0.0), self.alpha[jnp.where(self.alpha != 0.0)]
        alpha_terms = [f"alpha[{i[0]},{i[1]},{i[2]}]={float(v):.6f}" for i, v in zip(zip(*alpha_indices), alpha_value)]
        if len(alpha_terms) > 20:
            alpha_terms = alpha_terms[:20] + ["... (truncated)"]

        gamma_indices, gamma_value = jnp.where(self.gamma != 0.0), self.gamma[jnp.where(self.gamma != 0.0)]
        gamma_terms = [f"gamma[{i[0]},{i[1]},{i[2]},{i[3]}]={float(v):.6f}" for i, v in zip(zip(*gamma_indices), gamma_value)]
        if len(gamma_terms) > 20:
            gamma_terms = gamma_terms[:20] + ["... (truncated)"]

        return (f"NonlinearOscillator(n_modes={self.n_modes}, "
                f"{Q_terms}, {omega_0_terms}, "
                f"{', '.join(alpha_terms)}, {', '.join(gamma_terms)})")



from abc import ABC, abstractmethod

from .. import constants as const

class AbstractExcitation(ABC):
    def __init__(self):
        pass
    
    @abstractmethod
    def to_dtype(self, dtype):
        pass
    